#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009, 2012, 2013, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Test module for mpop.projector.
"""
import unittest
import numpy as np

from mock import MagicMock, patch
import sys
sys.modules['pyresample'] = MagicMock()
sys.modules['pyresample.bilinear'] = MagicMock()

from pyresample import geometry, utils

from mpop.projector import Projector
import mpop.projector


class TestProjector(unittest.TestCase):

    """Class for testing the Projector class.
    """

    proj = None

    @patch('mpop.projector.get_bil_info')
    @patch.object(utils, 'generate_quick_linesample_arrays')
    @patch.object(mpop.projector.kd_tree, 'get_neighbour_info')
    @patch.object(mpop.projector, '_get_area_hash')
    def test_init(self, gah, gni, gqla, bil_info):
        """Creation of coverage.
        """

        # in case of wrong number of arguments

        self.assertRaises(TypeError, Projector)
        self.assertRaises(TypeError, Projector, random_string(20))

        # in case of string arguments

        in_area_id = random_string(20)
        out_area_id = random_string(20)

        area_type = utils.parse_area_file.return_value.__getitem__.return_value

        gni.side_effect = [("a", "b", "c", "d")] * 10

        self.proj = Projector(in_area_id, out_area_id)
        self.assertEquals(utils.parse_area_file.call_count, 2)
        area_file = mpop.projector.get_area_file()
        utils.parse_area_file.assert_any_call(area_file, in_area_id)
        utils.parse_area_file.assert_any_call(area_file, out_area_id)

        self.assertEquals(self.proj.in_area, area_type)
        self.assertEquals(self.proj.out_area, area_type)

        # in case of undefined areas

        mock = MagicMock(side_effect=Exception("raise"))
        with patch.object(utils, 'parse_area_file', mock):
            self.assertRaises(Exception,
                              Projector,
                              "raise",
                              random_string(20))
            self.assertRaises(Exception,
                              Projector,
                              random_string(20),
                              "raise")

        # in case of geometry objects as input

        with patch.object(utils, 'AreaNotFound', Exception):
            mock = MagicMock(side_effect=[utils.AreaNotFound("raise"),
                                          MagicMock()])
            with patch.object(utils, 'parse_area_file', mock):
                in_area = geometry.AreaDefinition()
                self.proj = Projector(in_area, out_area_id)
                self.assertEquals(self.proj.in_area, in_area)

        in_area = geometry.SwathDefinition()
        utils.parse_area_file.return_value.__getitem__.side_effect = [
            AttributeError, out_area_id]
        self.proj = Projector(in_area, out_area_id)
        self.assertEquals(self.proj.in_area, in_area)

        out_area = geometry.AreaDefinition()
        utils.parse_area_file.return_value.__getitem__.side_effect = [
            in_area_id, AttributeError]
        self.proj = Projector(in_area_id, out_area)
        self.assertEquals(self.proj.out_area, out_area)

        # in case of lon/lat is input

        utils.parse_area_file.return_value.__getitem__.side_effect = [
            AttributeError, out_area_id]
        lonlats = ("great_lons", "even_greater_lats")

        self.proj = Projector("raise", out_area_id, lonlats)
        geometry.SwathDefinition.assert_called_with(lons=lonlats[0],
                                                    lats=lonlats[1])

        utils.parse_area_file.return_value.__getitem__.side_effect = None
        # in case of wrong mode

        self.assertRaises(ValueError,
                          Projector,
                          random_string(20),
                          random_string(20),
                          mode=random_string(20))

        utils.parse_area_file.return_value.__getitem__.side_effect = ["a", "b",
                                                                      "c", "d"]
        gqla.side_effect = [("ridx", "cidx")]
        # quick mode cache
        self.proj = Projector(in_area_id, out_area_id, mode="quick")
        cache = getattr(self.proj, "_cache")
        self.assertTrue(cache['row_idx'] is not None)
        self.assertTrue(cache['col_idx'] is not None)

        # nearest mode cache
        self.proj = Projector(in_area_id, out_area_id, mode="nearest")
        cache = getattr(self.proj, "_cache")
        self.assertTrue(cache['valid_index'] is not None)
        self.assertTrue(cache['valid_output_index'] is not None)
        self.assertTrue(cache['index_array'] is not None)

        # bilinear mode cache
        bil_info.return_value = (1, 2, 3, 4)

        def spam(val):
            return 'adef'

        with patch.object(mpop.projector, 'get_area_def', spam):
            self.proj = Projector(in_area_id, out_area_id, mode="bilinear")
        cache = getattr(self.proj, "_cache")
        self.assertTrue(cache['bilinear_t'] is not None)
        self.assertTrue(cache['bilinear_s'] is not None)
        self.assertTrue(cache['input_idxs'] is not None)
        self.assertTrue(cache['idx_arr'] is not None)

    @patch.object(np.ma, "array")
    @patch.object(mpop.projector.kd_tree, 'get_sample_from_neighbour_info')
    @patch.object(np, "load")
    def test_project_array(self, npload, gsfni, marray):
        """Test the project_array function.
        """
        in_area_id = random_string(20)
        out_area_id = random_string(20)
        data = np.random.standard_normal((3, 1))

        utils.parse_area_file.return_value.__getitem__.side_effect = [
            "a", "b", "c", "d"]
        # test quick
        self.proj = Projector(in_area_id, out_area_id, mode="quick")
        self.proj.project_array(data)
        mpop.projector.image.ImageContainer.assert_called_with(
            data, "a", fill_value=None)
        mpop.projector.image.ImageContainer.return_value.\
            get_array_from_linesample.assert_called_with(
                self.proj._cache["row_idx"], self.proj._cache["col_idx"])
        marray.assert_called_once_with(
            mpop.projector.image.ImageContainer.return_value.
            get_array_from_linesample.return_value,
            dtype=np.dtype('float64'))

        # test nearest
        in_area = MagicMock()
        out_area = MagicMock()
        utils.parse_area_file.return_value.__getitem__.side_effect = [
            in_area, out_area]
        self.proj = Projector(in_area_id, out_area_id, mode="nearest")
        self.proj.project_array(data)
        mpop.projector.kd_tree.get_sample_from_neighbour_info.\
            assert_called_with('nn',
                               out_area.shape,
                               data,
                               npload.return_value.__getitem__.return_value,
                               npload.return_value.__getitem__.return_value,
                               npload.return_value.__getitem__.return_value,
                               fill_value=None)

    @patch.object(mpop.projector.kd_tree, 'get_neighbour_info')
    def test_calc_nearest_params(self, gni):
        gni.return_value = (1, 2, 3, 4)
        res = mpop.projector.calc_nearest_params('in_area', 'out_area',
                                                 'radius', nprocs='nprocs')
        self.assertTrue(isinstance(res, dict))
        self.assertTrue('valid_index' in res)
        self.assertEqual(res['valid_index'], 1)
        self.assertTrue('valid_output_index' in res)
        self.assertEqual(res['valid_output_index'], 2)
        self.assertTrue('index_array' in res)
        self.assertEqual(res['index_array'], 3)

    @patch.object(mpop.projector.utils, 'generate_quick_linesample_arrays')
    def test_calc_quick_params(self, gqla):
        gqla.return_value = (1, 2)
        res = mpop.projector.calc_quick_params('in_area', 'out_area')

        self.assertTrue(isinstance(res, dict))
        self.assertTrue('row_idx' in res)
        self.assertEqual(res['row_idx'], 1)
        self.assertTrue('col_idx' in res)
        self.assertEqual(res['col_idx'], 2)

    @patch.object(mpop.projector, 'get_bil_info')
    def test_calc_bilinear_params(self, gbi):
        gbi.return_value = (1, 2, 3, 4)
        res = mpop.projector.calc_bilinear_params('in_area', 'out_area',
                                                  'radius', nprocs='nprocs')
        self.assertTrue(isinstance(res, dict))
        self.assertTrue('bilinear_t' in res)
        self.assertEqual(res['bilinear_t'], 1)
        self.assertTrue('bilinear_s' in res)
        self.assertEqual(res['bilinear_s'], 2)
        self.assertTrue('input_idxs' in res)
        self.assertEqual(res['input_idxs'], 3)
        self.assertTrue('idx_arr' in res)
        self.assertEqual(res['idx_arr'], 4)

    @patch.object(mpop.projector, 'll2cr')
    def test_calc_ewa_params(self, ll2):
        ll2.return_value = (0, 1, 2)
        res = mpop.projector.calc_ewa_params('in_area', 'out_area')
        self.assertTrue(isinstance(res, dict))
        self.assertTrue('ewa_cols' in res)
        self.assertEqual(res['ewa_cols'], 1)
        self.assertTrue('ewa_rows' in res)
        self.assertEqual(res['ewa_rows'], 2)

    def test_get_precompute_cache_fname(self):
        res = mpop.projector.get_precompute_cache_fname('in_id', 'out_id',
                                                        'in_area', 'out_area',
                                                        'mode', 'proj_dir')
        cor_res = "proj_dir/in_id2out_id_-" + \
                  "6296787761359943868to8984161303220364208_mode.npz"
        self.assertTrue(res == cor_res)

    @patch.object(mpop.projector, 'get_area_def')
    @patch.object(mpop.projector.geometry, 'SwathDefinition')
    def test_get_area_and_id(self, swath_def, gad):
        # Case when get_area_def works
        swath_def.return_value = 1
        gad.return_value = 'adef'
        res = mpop.projector.get_area_and_id('area')
        self.assertTrue(res[0] == 'adef')
        self.assertTrue(res[1] == 'area')
        # Case when AttributeError is raised
        with self.assertRaises(AttributeError):
            gad.side_effect = AttributeError
            res = mpop.projector.get_area_and_id('area')
        # Case when AttributeError is raised and latlons are given
        gad.side_effect = AttributeError
        res = mpop.projector.get_area_and_id('area', latlons=[1, 2])
        self.assertEqual(res[0], 1)
        self.assertTrue(res[1], 'area')


def random_string(length,
                  choices="abcdefghijklmnopqrstuvwxyz"
                  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
    """Generates a random string with elements from *set* of the specified
    *length*.
    """
    import random
    return "".join([random.choice(choices)
                    for dummy in range(length)])


def suite():
    """The test suite for test_projector.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestProjector))

    return mysuite



# Copyright (c) 2012, 2015
#

# Author(s):
#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
Convert human coordinates (lon and lat) to an area definition.

Here is a usage example:
python coord2area_def.py france stere 42.0 51.5 -5.5 8.0 1.5
(the arguments are "name proj min_lat max_lat min_lon max_lon resolution(km)")


and the result is:
REGION: france {
	NAME:	france
	PCS_ID:	stere_1.25_46.75
	PCS_DEF:	proj=stere,lat_0=46.75,lon_0=1.25,ellps=WGS84
	XSIZE:  746
	YSIZE:  703
	AREA_EXTENT:	(-559750.38109755167, -505020.6757764442,
559750.38109755167, 549517.35194826045)
};

"""

import sys
from pyproj import Proj
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("name",
                        help="The name of the area.")
    parser.add_argument("proj",
                        help="The projection to use. Use proj.4 names, like 'stere', 'merc'...")
    parser.add_argument("min_lat",
                        help="The the latitude of the bottom of the area",
                        type=float)
    parser.add_argument("max_lat",
                        help="The the latitude of the top of the area",
                        type=float)
    parser.add_argument("min_lon",
                        help="The the longitude of the left of the area",
                        type=float)
    parser.add_argument("max_lon",
                        help="The the longitude of the right of the area",
                        type=float)
    parser.add_argument("resolution",
                        help="The resolution of the area (in km)",
                        type=float)
    parser.add_argument("-s", "--shapes",
                        help="Show a preview of the area using the coastlines in this directory")

    args = parser.parse_args()
    name = args.name
    proj = args.proj

    left = args.min_lon
    right = args.max_lon
    up = args.min_lat
    down = args.max_lat

    res = args.resolution * 1000

    lat_0 = (up + down) / 2
    lon_0 = (right + left) / 2

    p = Proj(proj=proj, lat_0=lat_0, lon_0=lon_0, ellps="WGS84")

    left_ex1, up_ex1 = p(left, up)
    right_ex1, up_ex2 = p(right, up)
    left_ex2, down_ex1 = p(left, down)
    right_ex2, down_ex2 = p(right, down)
    left_ex3, dummy = p(left, lat_0)
    right_ex3, dummy = p(right, lat_0)

    area_extent = (min(left_ex1, left_ex2, left_ex3),
                   min(up_ex1, up_ex2),
                   max(right_ex1, right_ex2, right_ex3),
                   max(down_ex1, down_ex2))

    xsize = int((area_extent[2] - area_extent[0]) / res)
    ysize = int((area_extent[3] - area_extent[1]) / res)

    proj4_string = "+" + \
        " +".join(("proj=" + proj + ",lat_0=" + str(lat_0) + ",lon_0=" + str(lon_0) + ",ellps=WGS84").split(","))

    print proj4_string

    print "REGION:", name, "{"
    print "\tNAME:\t", name
    print "\tPCS_ID:\t", proj + "_" + str(lon_0) + "_" + str(lat_0)
    print ("\tPCS_DEF:\tproj=" + proj +
           ",lat_0=" + str(lat_0) +
           ",lon_0=" + str(lon_0) +
           ",ellps=WGS84")
    print "\tXSIZE:\t", xsize
    print "\tYSIZE:\t", ysize
    print "\tAREA_EXTENT:\t", area_extent
    print "};"

    if args.shapes is None:
        sys.exit(0)
    from PIL import Image
    from pycoast import ContourWriterAGG
    img = Image.new('RGB', (xsize, ysize))
    #proj4_string = '+proj=geos +lon_0=0.0 +a=6378169.00 +b=6356583.80 +h=35785831.0'
    #area_extent = (-5570248.4773392612, -5567248.074173444, 5567248.074173444, 5570248.4773392612)
    area_def = (proj4_string, area_extent)
    cw = ContourWriterAGG(args.shapes)
    #cw = ContourWriterAGG('/usr/share/gshhg-gmt-shp/')
    cw.add_coastlines(img, (proj4_string, area_extent), resolution='l', width=0.5)

    cw.add_grid(img, area_def, (10.0, 10.0), (2.0, 2.0), write_text=False, outline='white', outline_opacity=175, width=1.0,
                minor_outline='white', minor_outline_opacity=175, minor_width=0.2, minor_is_tick=False)
    img.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2014, 2015
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This file is part of mpop.
#
# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

'''Helper functions for eg. performing Sun zenith angle correction.
'''

import numpy as np


def sunzen_corr_cos(data, cos_zen, limit=80.):
    '''Perform Sun zenith angle correction to the given *data* using
    cosine of the zenith angle (*cos_zen*).  The correction is limited
    to *limit* degrees (default: 80.0 degrees).  For larger zenith
    angles, the correction is the same as at the *limit*.  Both *data*
    and *cos_zen* are given as 2-dimensional Numpy arrays or Numpy
    MaskedArrays, and they should have equal shapes.
    '''

    # Convert the zenith angle limit to cosine of zenith angle
    cos_limit = np.cos(np.radians(limit))

    # Cosine correction
    lim_y, lim_x = np.where(cos_zen > cos_limit)
    data[lim_y, lim_x] /= cos_zen[lim_y, lim_x]
    # Use constant value (the limit) for larger zenith
    # angles
    lim_y, lim_x = np.where(cos_zen <= cos_limit)
    data[lim_y, lim_x] /= cos_limit

    return data


def estimate_cth(IR_108, cth_atm="standard"):

    '''
    Estimation of the cloud top height using the 10.8 micron channel
    limitations: this is the most simple approach
                 a simple fit of the ir108 to the temperature profile
                 * no correction for water vapour or any other trace gas
                 * no viewing angle dependency
                 * no correction for semi-transparent clouds

    optional input:
      cth_atm    * "standard", "tropics", "midlatitude summer", "midlatitude winter", "subarctic summer", "subarctic winter"
                  Matching the 10.8 micron temperature with atmosphere profile
                  (s)  AFGL atmospheric constituent profile. U.S. standard atmosphere 1976. (AFGL-TR-86-0110) 
                  (t)  AFGL atmospheric constituent profile. tropical.                      (AFGL-TR-86-0110)
                  (mw) AFGL atmospheric constituent profile. midlatitude summer.            (AFGL-TR-86-0110) 
                  (ms) AFGL atmospheric constituent profile. midlatitude winter.            (AFGL-TR-86-0110)
                  (ss) AFGL atmospheric constituent profile. subarctic summer.              (AFGL-TR-86-0110) 
                  (sw) AFGL atmospheric constituent profile. subarctic winter.              (AFGL-TR-86-0110)
                  Ulrich Hamann (MeteoSwiss)
                * "tropopause"
                  Assuming a fixed tropopause height and a fixed temperature gradient
                  Richard Mueller (DWD)
    output: 
      parallax corrected channel
                 the content of the channel will be parallax corrected.
                 The name of the new channel will be
                 *original_chan.name+'_PC'*, eg. "IR_108_PC". This name is
                 also stored to the info dictionary of the originating channel.

    Versions: 05.07.2016 initial version
              Ulrich Hamann (MeteoSwiss), Richard Mueller (DWD)
    '''

    print "*** estimating CTH using the 10.8 micro meter brightness temperature "

    if cth_atm.lower() != "tropopause":

        # define atmospheric temperature profile    
        import os
        from numpy import loadtxt, zeros, where, logical_and
        import mpop 

        mpop_dir = os.path.dirname(mpop.__file__)
        afgl_file = mpop_dir+"/afgl.dat"
        print "... assume ", cth_atm, " atmosphere for temperature profile"

        if cth_atm.lower()=="standard" or cth_atm.lower()=="s":
            z, T = loadtxt(afgl_file, usecols=(0, 1), unpack=True, comments="#")
        elif cth_atm.lower()=="tropics" or cth_atm.lower()=="t":
            z, T = loadtxt(afgl_file, usecols=(0, 2), unpack=True, comments="#")
        elif cth_atm.lower()=="midlatitude summer" or cth_atm.lower()=="ms":
            z, T = loadtxt(afgl_file, usecols=(0, 3), unpack=True, comments="#")
        elif cth_atm.lower()=="midlatitude winter" or cth_atm.lower()=="ws":
            z, T = loadtxt(afgl_file, usecols=(0, 4), unpack=True, comments="#")
        elif cth_atm.lower()=="subarctic summer" or cth_atm.lower()=="ss":
            z, T = loadtxt(afgl_file, usecols=(0, 5), unpack=True, comments="#")
        elif cth_atm.lower()=="subarctic winter" or cth_atm.lower()=="ss":
            z, T = loadtxt(afgl_file, usecols=(0, 6), unpack=True, comments="#")
        else:
            print "*** Error in estimate_cth (mpop/tools.py)"
            print "unknown temperature profiel for CTH estimation: cth_atm = ", cth_atm
            quit()

        height = zeros(IR_108.shape)
        # warmer than lowest level -> clear sky 
        height[where(IR_108 > T[-1])] = -1.
        print "     z0(km)   z1(km)   T0(K)   T1(K)  number of pixels"
        print "------------------------------------------------------"
        for i in range(z.size)[::-1]:

            # search for temperatures between layer i-1 and i
            ind =  np.where( logical_and( T[i-1]< IR_108, IR_108 < T[i]) )
            # interpolate CTH according to ir108 temperature
            height[ind] = z[i] + (IR_108[ind]-T[i])/(T[i-1]-T[i]) * (z[i-1]-z[i])
            # verbose output
            print " {0:8.1f} {1:8.1f} {2:8.1f} {3:8.1f} {4:8d}".format(z[i], z[i-1], T[i], T[i-1], len(ind[0]))

            # if temperature increases above 8km -> tropopause detected
            if z[i]>=8. and T[i] <= T[i-1]:
                # no cloud above tropopose
                break
            # no cloud heights above 20km
            if z[i]>=20.:
                break

        # if height is still 0 -> cloud colder than tropopause -> cth == tropopause height
        height[np.where( height == 0 )] = z[i]
        
    else:

        Htropo=11.0 # km
        # this is an assumption it should be optimized 
        # by making it dependent on region and season. 
        # It might be good to include the ITC in the  
        # region of interest, that would make a fixed Htropo 
        # value more reliable. 
        Tmin = np.amin(IR_108) 
        # for Tmin it might be better to use the 5th or 10th percentile 
        # else overshoting tops induces further uncertainties  
        # in the calculation of the cloud height. 
        # However numpy provides weird results for 5th percentile. 
        # Hence, for the working version the minima is used 

        print "... assume tropopause height ", Htropo, ", tropopause temperature ", Tmin, "K (", Tmin-273.16, "deg C)"
        print "    and constant temperature gradient 6.5 K/km"

        height = -(IR_108 - Tmin)/6.5 + Htropo 
        # calculation of the height, the temperature gradient 
        # 6.5 K/km is an assumption  
        # derived from USS and MPI standard profiles. It 
        # has to be improved as well 

    # convert to masked array
    # convert form km to meter
    height = np.ma.masked_where(height <= 0, height, copy=False) * 1000.

    if False:
        from trollimage.image import Image as trollimage
        from trollimage.colormap import rainbow
        from copy import deepcopy 
        # cloud top height
        prop = height
        min_data = prop.min()
        max_data = prop.max()
        print " estimated CTH(meter) (min/max): ", min_data, max_data
        min_data =     0
        max_data = 12000    
        colormap = deepcopy(rainbow)
        colormap.set_range(min_data, max_data)
        img = trollimage(prop, mode="L") #, fill_value=[0,0,0]
        img.colorize(colormap)
        img.show()

    # return cloud top height in meter
    return height


def viewzen_corr(data, view_zen):
    """Apply atmospheric correction on the given *data* using the
    specified satellite zenith angles (*view_zen*). Both input data
    are given as 2-dimensional Numpy (masked) arrays, and they should
    have equal shapes.
    The *data* array will be changed in place and has to be copied before.
    """
    def ratio(value, v_null, v_ref):
        return (value - v_null) / (v_ref - v_null)

    def tau0(t):
        T_0 = 210.0
        T_REF = 320.0
        TAU_REF = 9.85
        return (1 + TAU_REF)**ratio(t, T_0, T_REF) - 1

    def tau(t):
        T_0 = 170.0
        T_REF = 295.0
        TAU_REF = 1.0
        M = 4
        return TAU_REF * ratio(t, T_0, T_REF)**M

    def delta(z):
        Z_0 = 0.0
        Z_REF = 70.0
        DELTA_REF = 6.2
        return (1 + DELTA_REF)**ratio(z, Z_0, Z_REF) - 1

    y0, x0 = np.ma.where(view_zen == 0)
    data[y0, x0] += tau0(data[y0, x0])

    y, x = np.ma.where((view_zen > 0) & (view_zen < 90) & (~data.mask))
    data[y, x] += tau(data[y, x]) * delta(view_zen[y, x])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012, 2013, 2014, 2015, 2016, 2017 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Nina Håkansson <nina.hakansson@smhi.se>
#   Oana Nicola <oananicola@yahoo.com>
#   Lars Ørum Rasmussen <ras@dmi.dk>
#   Panu Lahtinen <panu.lahtinen@fmi.fi>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for aapp level 1b data.

Options for loading:

 - pre_launch_coeffs (False): use pre-launch coefficients if True, operational
   otherwise (if available).

http://research.metoffice.gov.uk/research/interproj/nwpsaf/aapp/
NWPSAF-MF-UD-003_Formats.pdf
"""


import numpy as np
import os
import logging
import datetime
import glob
from ConfigParser import ConfigParser
from mpop import CONFIG_PATH

LOGGER = logging.getLogger('aapp1b')


def load(satscene, *args, **kwargs):
    """Read data from file and load it into *satscene*.
    A possible *calibrate* keyword argument is passed to the AAPP reader.
    Should be 0 for off (counts), 1 for default (brightness temperatures and
    reflectances), and 2 for radiances only.

    If *use_extern_calib* keyword argument is set True, use external
    calibration data.

    """
    del args

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value

    if kwargs.get("filename") is not None:
        options["full_filename"] = kwargs["filename"]

    options["calibrate"] = kwargs.get("calibrate", True)
    options["pre_launch_coeffs"] = kwargs.get("pre_launch_coeffs", False)
    options["use_extern_calib"] = kwargs.get("use_extern_calib", False)

    LOGGER.info("Loading instrument '%s'", satscene.instrument_name)

    try:
        CASES[satscene.instrument_name](satscene, options)
    except KeyError:
        raise KeyError("Unknown instrument '%s'" % satscene.instrument_name)


def load_avhrr(satscene, options):
    """Read avhrr data from file and load it into *satscene*.
    """
    if "filename" not in options:
        raise IOError("No filename given, cannot load.")

    loaded = set([chn.name for chn in satscene.loaded_channels()])

    chns = (satscene.channels_to_load &
            (set(AVHRR_CHANNEL_NAMES) - loaded))

    LOGGER.info("Loading channels %s", str(sorted(list(chns))))

    if len(chns) == 0:
        return

    values = {"orbit": '%05d' % int(satscene.orbit),
              "satname": satscene.satname,
              "number": satscene.number,
              "instrument": satscene.instrument_name,
              "satellite": satscene.fullname
              }

    done_reading = False

    if "full_filename" in options:
        filename = options["full_filename"]
        LOGGER.debug("Loading from %s", filename)
        scene = AAPP1b(filename)
        try:
            scene.read()
            done_reading = True
        except ValueError:
            LOGGER.info("Can't read %s", filename)

    if not done_reading:
        filename = \
            os.path.join(satscene.time_slot.strftime(options["dir"]) % values,
                         satscene.time_slot.strftime(
                             options["filename"])
                         % values)

        file_list = glob.glob(filename)

        if len(file_list) > 1:
            LOGGER.info("More than one l1b file found: %s", str(file_list))
            # hrpt_noaa18_20150110_1658_49685.l1b
            candidate = ('hrpt_' +
                         str(satscene.satname) + str(satscene.number) +
                         satscene.time_slot.strftime('_%Y%m%d_%H%M_') +
                         '%05d' % int(satscene.orbit) + '.l1b')
            LOGGER.debug("Suggested filename = %s", str(candidate))
            candidate_found = False
            for fname in file_list:
                l1bname = os.path.basename(fname)
                if l1bname == candidate:
                    filename = fname
                    candidate_found = True
                    LOGGER.info('The l1b file chosen is this: %s',
                                str(filename))
                    break
            if not candidate_found:
                LOGGER.info("More than one l1b file found: %s", str(file_list))
                LOGGER.warning("Couldn't decide which one to take. "
                               "Try take the first one: %s", str(filename))
                filename = file_list[0]

        elif len(file_list) == 0:
            raise IOError("No l1b file matching!: %s", filename)
        else:
            filename = file_list[0]

        LOGGER.debug("Loading from %s", filename)
        scene = AAPP1b(filename)
        try:
            scene.read()
        except ValueError:
            LOGGER.info("Can't read %s, exiting.", filename)
            return

    calib_coeffs = None
    if options["use_extern_calib"]:
        import h5py
        LOGGER.info("Reading external calibration coefficients.")
        try:
            fid = h5py.File(os.path.join(CONFIG_PATH, satscene.satname +
                                         '_calibration_data.h5'), 'r')
            calib_coeffs = {}
            for key in fid.keys():
                date_diffs = []
                for dat in fid[key]['datetime']:
                    date_diffs.append(np.abs(satscene.time_slot -
                                             datetime.datetime(dat[0],
                                                               dat[1],
                                                               dat[2])))
                idx = date_diffs.index(min(date_diffs))
                date_diff = satscene.time_slot - \
                    datetime.datetime(fid[key]['datetime'][idx][0],
                                      fid[key]['datetime'][idx][1],
                                      fid[key]['datetime'][idx][2])
                if date_diff.days < 0:
                    older_or_newer = "newer"
                else:
                    older_or_newer = "older"
                LOGGER.info("External calibration for %s is %d "
                            "days %s than data.",
                            key, date_diffs[idx].days, older_or_newer)
                calib_coeffs[key] = (fid[key]['slope1'][idx],
                                     fid[key]['intercept1'][idx],
                                     fid[key]['slope2'][idx],
                                     fid[key]['intercept2'][idx])
            fid.close()

            if 'ch1' not in calib_coeffs:
                calib_coeffs['ch1'] = None
            if 'ch2' not in calib_coeffs:
                calib_coeffs['ch2'] = None
            if 'ch3a' not in calib_coeffs:
                calib_coeffs['ch3a'] = None

        except IOError:
            LOGGER.info("No external calibration data found.")

    scene.calibrate(chns, calibrate=options.get('calibrate', 1),
                    pre_launch_coeffs=options["pre_launch_coeffs"],
                    calib_coeffs=calib_coeffs)

    if satscene.area is None:
        scene.navigate()

        try:
            from pyresample import geometry
        except ImportError as ex_:

            LOGGER.debug("Could not load pyresample: %s", str(ex_))

            satscene.lat = scene.lats
            satscene.lon = scene.lons
        else:
            satscene.area = geometry.SwathDefinition(lons=scene.lons,
                                                     lats=scene.lats)
            area_name = ("swath_" + satscene.fullname + "_" +
                         str(satscene.time_slot) + "_" +
                         str(scene.lats.shape))
            satscene.area.area_id = area_name
            satscene.area.name = "Satellite projection"
            satscene.area_id = area_name

    for chn in chns:
        if (chn in scene.channels) and np.ma.count(scene.channels[chn]) > 0:
            satscene[chn].data = scene.channels[chn]
            satscene[chn].info['units'] = scene.units[chn]
            satscene[chn].area = satscene.area
        else:
            del satscene[chn]


AVHRR_CHANNEL_NAMES = ("1", "2", "3A", "3B", "4", "5")

# AAPP 1b header

_HEADERTYPE = np.dtype([("siteid", "S3"),
                        ("blank", "S1"),
                        ("l1bversnb", "<i2"),
                        ("l1bversyr", "<i2"),
                        ("l1bversdy", "<i2"),
                        ("reclg", "<i2"),
                        ("blksz", "<i2"),
                        ("hdrcnt", "<i2"),
                        ("filler0", "S6"),
                        ("dataname", "S42"),
                        ("prblkid", "S8"),
                        ("satid", "<i2"),
                        ("instid", "<i2"),
                        ("datatype", "<i2"),
                        ("tipsrc", "<i2"),
                        ("startdatajd", "<i4"),
                        ("startdatayr", "<i2"),
                        ("startdatady", "<i2"),
                        ("startdatatime", "<i4"),
                        ("enddatajd", "<i4"),
                        ("enddatayr", "<i2"),
                        ("enddatady", "<i2"),
                        ("enddatatime", "<i4"),
                        ("cpidsyr", "<i2"),
                        ("cpidsdy", "<i2"),
                        ("filler1", "S8"),
                        # data set quality indicators
                        ("inststat1", "<i4"),
                        ("filler2", "S2"),
                        ("statchrecnb", "<i2"),
                        ("inststat2", "<i4"),
                        ("scnlin", "<i2"),
                        ("callocscnlin", "<i2"),
                        ("misscnlin", "<i2"),
                        ("datagaps", "<i2"),
                        ("okdatafr", "<i2"),
                        ("pacsparityerr", "<i2"),
                        ("auxsyncerrsum", "<i2"),
                        ("timeseqerr", "<i2"),
                        ("timeseqerrcode", "<i2"),
                        ("socclockupind", "<i2"),
                        ("locerrind", "<i2"),
                        ("locerrcode", "<i2"),
                        ("pacsstatfield", "<i2"),
                        ("pacsdatasrc", "<i2"),
                        ("filler3", "S4"),
                        ("spare1", "S8"),
                        ("spare2", "S8"),
                        ("filler4", "S10"),
                        # Calibration
                        ("racalind", "<i2"),
                        ("solarcalyr", "<i2"),
                        ("solarcaldy", "<i2"),
                        ("pcalalgind", "<i2"),
                        ("pcalalgopt", "<i2"),
                        ("scalalgind", "<i2"),
                        ("scalalgopt", "<i2"),
                        ("irttcoef", "<i2", (4, 6)),
                        ("filler5", "<i4", (2, )),
                        # radiance to temperature conversion
                        ("albcnv", "<i4", (2, 3)),
                        ("radtempcnv", "<i4", (3, 3)),
                        ("filler6", "<i4", (3, )),
                        # Navigation
                        ("modelid", "S8"),
                        ("nadloctol", "<i2"),
                        ("locbit", "<i2"),
                        ("filler7", "S2"),
                        ("rollerr", "<i2"),
                        ("pitcherr", "<i2"),
                        ("yawerr", "<i2"),
                        ("epoyr", "<i2"),
                        ("epody", "<i2"),
                        ("epotime", "<i4"),
                        ("smaxis", "<i4"),
                        ("eccen", "<i4"),
                        ("incli", "<i4"),
                        ("argper", "<i4"),
                        ("rascnod", "<i4"),
                        ("manom", "<i4"),
                        ("xpos", "<i4"),
                        ("ypos", "<i4"),
                        ("zpos", "<i4"),
                        ("xvel", "<i4"),
                        ("yvel", "<i4"),
                        ("zvel", "<i4"),
                        ("earthsun", "<i4"),
                        ("filler8", "S16"),
                        # analog telemetry conversion
                        ("pchtemp", "<i2", (5, )),
                        ("reserved1", "<i2"),
                        ("pchtempext", "<i2", (5, )),
                        ("reserved2", "<i2"),
                        ("pchpow", "<i2", (5, )),
                        ("reserved3", "<i2"),
                        ("rdtemp", "<i2", (5, )),
                        ("reserved4", "<i2"),
                        ("bbtemp1", "<i2", (5, )),
                        ("reserved5", "<i2"),
                        ("bbtemp2", "<i2", (5, )),
                        ("reserved6", "<i2"),
                        ("bbtemp3", "<i2", (5, )),
                        ("reserved7", "<i2"),
                        ("bbtemp4", "<i2", (5, )),
                        ("reserved8", "<i2"),
                        ("eleccur", "<i2", (5, )),
                        ("reserved9", "<i2"),
                        ("motorcur", "<i2", (5, )),
                        ("reserved10", "<i2"),
                        ("earthpos", "<i2", (5, )),
                        ("reserved11", "<i2"),
                        ("electemp", "<i2", (5, )),
                        ("reserved12", "<i2"),
                        ("chtemp", "<i2", (5, )),
                        ("reserved13", "<i2"),
                        ("bptemp", "<i2", (5, )),
                        ("reserved14", "<i2"),
                        ("mhtemp", "<i2", (5, )),
                        ("reserved15", "<i2"),
                        ("adcontemp", "<i2", (5, )),
                        ("reserved16", "<i2"),
                        ("d4bvolt", "<i2", (5, )),
                        ("reserved17", "<i2"),
                        ("d5bvolt", "<i2", (5, )),
                        ("reserved18", "<i2"),
                        ("bbtempchn3B", "<i2", (5, )),
                        ("reserved19", "<i2"),
                        ("bbtempchn4", "<i2", (5, )),
                        ("reserved20", "<i2"),
                        ("bbtempchn5", "<i2", (5, )),
                        ("reserved21", "<i2"),
                        ("refvolt", "<i2", (5, )),
                        ("reserved22", "<i2"),
                        ])

# AAPP 1b scanline

_SCANTYPE = np.dtype([("scnlin", "<i2"),
                      ("scnlinyr", "<i2"),
                      ("scnlindy", "<i2"),
                      ("clockdrift", "<i2"),
                      ("scnlintime", "<i4"),
                      ("scnlinbit", "<i2"),
                      ("filler0", "S10"),
                      ("qualind", "<i4"),
                      ("scnlinqual", "<i4"),
                      ("calqual", "<i2", (3, )),
                      ("cbiterr", "<i2"),
                      ("filler1", "S8"),
                      # Calibration
                      ("calvis", "<i4", (3, 3, 5)),
                      ("calir", "<i4", (3, 2, 3)),
                      ("filler2", "<i4", (3, )),
                      # Navigation
                      ("navstat", "<i4"),
                      ("attangtime", "<i4"),
                      ("rollang", "<i2"),
                      ("pitchang", "<i2"),
                      ("yawang", "<i2"),
                      ("scalti", "<i2"),
                      ("ang", "<i2", (51, 3)),
                      ("filler3", "<i2", (3, )),
                      ("pos", "<i4", (51, 2)),
                      ("filler4", "<i4", (2, )),
                      ("telem", "<i2", (103, )),
                      ("filler5", "<i2"),
                      ("hrpt", "<i2", (2048, 5)),
                      ("filler6", "<i4", (2, )),
                      # tip minor frame header
                      ("tipmfhd", "<i2", (7, 5)),
                      # cpu telemetry
                      ("cputel", "S6", (2, 5)),
                      ("filler7", "<i2", (67, )),
                      ])


class AAPP1b(object):

    """AAPP-level 1b data reader
    """

    def __init__(self, fname):
        if not isinstance(fname, (list, tuple, set)):
            fname = [fname]
        self.filenames = fname
        self.channels = dict([(i, None) for i in AVHRR_CHANNEL_NAMES])
        self.units = dict([(i, 'counts') for i in AVHRR_CHANNEL_NAMES])

        self._data = None
        self._header = None
        self._is3b = None
        self.lons = None
        self.lats = None

    def read(self):
        """Read the data.
        """
        tic = datetime.datetime.now()
        all_data = []
        all_headers = []
        for fname in self.filenames:
            with open(fname, "rb") as fp_:
                header = np.memmap(fp_, dtype=_HEADERTYPE, mode="r",
                                   shape=(_HEADERTYPE.itemsize, ))
                all_headers.append(header)
                data = np.memmap(fp_,
                                 dtype=_SCANTYPE,
                                 offset=22016, mode="r")
                all_data.append(data)

        LOGGER.debug("Reading time %s", str(datetime.datetime.now() - tic))

        self._header = all_headers
        self._data = all_data

    def navigate(self):
        """Return the longitudes and latitudes of the scene.
        """
        tic = datetime.datetime.now()
        lons40km, lats40km = [], []
        for data in self._data:
            lons40km.append(data["pos"][:, :, 1] * 1e-4)
            lats40km.append(data["pos"][:, :, 0] * 1e-4)

        lons40km = np.vstack(lons40km)
        lats40km = np.vstack(lats40km)

        try:
            from geotiepoints import SatelliteInterpolator
        except ImportError:
            LOGGER.warning("Could not interpolate lon/lats, "
                           "python-geotiepoints missing.")
            self.lons, self.lats = lons40km, lats40km
        else:
            cols40km = np.arange(24, 2048, 40)
            cols1km = np.arange(2048)
            lines = lons40km.shape[0]
            rows40km = np.arange(lines)
            rows1km = np.arange(lines)

            along_track_order = 1
            cross_track_order = 3

            satint = SatelliteInterpolator((lons40km, lats40km),
                                           (rows40km, cols40km),
                                           (rows1km, cols1km),
                                           along_track_order,
                                           cross_track_order)
            self.lons, self.lats = satint.interpolate()
            LOGGER.debug("Navigation time %s",
                         str(datetime.datetime.now() - tic))

    def calibrate(self, chns=("1", "2", "3A", "3B", "4", "5"),
                  calibrate=1, pre_launch_coeffs=False, calib_coeffs=None):
        """Calibrate the data
        """
        tic = datetime.datetime.now()

        channels = {'1': [], '2': [], '3A': [], '3B': [], '4': [], '5': []}
        is3b_all = []

        i = 0
        for data in self._data:

            if "1" in chns:
                LOGGER.debug("Calibrating channel 1.")
                if calib_coeffs is not None:
                    coeffs = calib_coeffs['ch1']
                else:
                    coeffs = None
                channels['1'].append(_vis_calibrate(data, 0,
                                                    calibrate,
                                                    pre_launch_coeffs,
                                                    coeffs))
                self.units['1'] = '%'
                if calibrate == 0:
                    self.units['1'] = ''

            if "2" in chns:
                LOGGER.debug("Calibrating channel 2.")
                if calib_coeffs is not None:
                    coeffs = calib_coeffs['ch2']
                else:
                    coeffs = None
                channels['2'].append(_vis_calibrate(data, 1,
                                                    calibrate,
                                                    pre_launch_coeffs,
                                                    coeffs))
                self.units['2'] = '%'
                if calibrate == 0:
                    self.units['2'] = ''

            if "3A" in chns or "3B" in chns:
                # Is it 3A or 3B:
                is3b = np.expand_dims(np.bitwise_and(
                    np.right_shift(data['scnlinbit'], 0), 1) == 1, 1)
                # self._is3b = is3b
                is3b_all.append(is3b)

            if "3A" in chns:
                LOGGER.debug("Calibrating channel 3a.")
                if calib_coeffs is not None:
                    coeffs = calib_coeffs['ch3a']
                else:
                    coeffs = None
                ch3a = _vis_calibrate(data, 2,
                                      calibrate, pre_launch_coeffs,
                                      coeffs)
                channels['3A'].append(np.ma.masked_array(ch3a, is3b * ch3a))

                self.units['3A'] = '%'
                if calibrate == 0:
                    self.units['3A'] = ''

            if "3B" in chns:
                LOGGER.debug("Calibrating channel 3b.")
                ch3b = _ir_calibrate(self._header[i], data, 0, calibrate)
                channels['3B'].append(
                    np.ma.masked_array(ch3b,
                                       np.logical_or((is3b is False) *
                                                     ch3b,
                                                     ch3b < 0.1)))
                if calibrate == 1:
                    self.units['3B'] = 'K'
                elif calibrate == 2:
                    self.units['3B'] = 'W*m-2*sr-1*cm ?'
                else:
                    self.units['3B'] = ''

            if "4" in chns:
                LOGGER.debug("Calibrating channel 4.")
                channels['4'].append(_ir_calibrate(self._header[i],
                                                   data, 1, calibrate))
                if calibrate == 1:
                    self.units['4'] = 'K'
                elif calibrate == 2:
                    self.units['4'] = 'W*m-2*sr-1*cm ?'
                else:
                    self.units['4'] = ''

            if "5" in chns:
                LOGGER.debug("Calibrating channel 5.")
                channels['5'].append(_ir_calibrate(self._header[i],
                                                   data, 2, calibrate))
                if calibrate == 1:
                    self.units['5'] = 'K'
                elif calibrate == 2:
                    self.units['5'] = 'W*m-2*sr-1*cm ?'
                else:
                    self.units['5'] = ''

            i += 1

        # transfer channel data to class attributes
        for ch_ in channels:
            try:
                self.channels[ch_] = np.ma.vstack(channels[ch_])
            except ValueError:
                self.channels[ch_] = None

        LOGGER.debug("Calibration time %s", str(datetime.datetime.now() - tic))


def _vis_calibrate(data, chn, calib_type, pre_launch_coeffs=False,
                   calib_coeffs=None):
    """Visible channel calibration only.
    *calib_type* = 0: Counts
    *calib_type* = 1: Reflectances
    *calib_type* = 2: Radiances
    """
    # Calibration count to albedo, the calibration is performed separately for
    # two value ranges.

    channel = data["hrpt"][:, :, chn].astype(np.float)
    if calib_type == 0:
        return channel

    if calib_type == 2:
        LOGGER.info("Radiances are not yet supported for " +
                    "the VIS/NIR channels!")

    if pre_launch_coeffs:
        coeff_idx = 2
    else:
        # check that coeffs are valid
        if np.all(data["calvis"][:, chn, 0, 4] == 0):
            LOGGER.info(
                "No valid operational coefficients, fall back to pre-launch")
            coeff_idx = 2
        else:
            coeff_idx = 0

    intersection = data["calvis"][:, chn, coeff_idx, 4]

    if calib_coeffs is not None:
        LOGGER.info("Updating from external calibration coefficients.")
        # intersection = np.expand_dims
        slope1 = np.expand_dims(calib_coeffs[0], 1)
        intercept1 = np.expand_dims(calib_coeffs[1], 1)
        slope2 = np.expand_dims(calib_coeffs[2], 1)
        intercept2 = np.expand_dims(calib_coeffs[3], 1)
    else:
        slope1 = \
            np.expand_dims(data["calvis"][:, chn, coeff_idx, 0] * 1e-10, 1)
        intercept1 = \
            np.expand_dims(data["calvis"][:, chn, coeff_idx, 1] * 1e-7, 1)
        slope2 = \
            np.expand_dims(data["calvis"][:, chn, coeff_idx, 2] * 1e-10, 1)
        intercept2 = \
            np.expand_dims(data["calvis"][:, chn, coeff_idx, 3] * 1e-7, 1)

        if chn == 2:
            slope2[slope2 < 0] += 0.4294967296

    mask1 = channel <= np.expand_dims(intersection, 1)
    mask2 = channel > np.expand_dims(intersection, 1)

    channel[mask1] = (channel * slope1 + intercept1)[mask1]

    channel[mask2] = (channel * slope2 + intercept2)[mask2]

    channel[channel < 0] = np.nan
    return np.ma.masked_invalid(channel)


def _ir_calibrate(header, data, irchn, calib_type):
    """IR calibration
    *calib_type* = 0: Counts
    *calib_type* = 1: BT
    *calib_type* = 2: Radiances
    """

    count = data['hrpt'][:, :, irchn + 2].astype(np.float)

    if calib_type == 0:
        return count

    k1_ = np.expand_dims(data['calir'][:, irchn, 0, 0] / 1.0e9, 1)
    k2_ = np.expand_dims(data['calir'][:, irchn, 0, 1] / 1.0e6, 1)
    k3_ = np.expand_dims(data['calir'][:, irchn, 0, 2] / 1.0e6, 1)

    # Count to radiance conversion:
    rad = k1_ * count * count + k2_ * count + k3_

    all_zero = np.logical_and(np.logical_and(np.equal(k1_, 0),
                                             np.equal(k2_, 0)),
                              np.equal(k3_, 0))
    idx = np.indices((all_zero.shape[0],))
    suspect_line_nums = np.repeat(idx[0], all_zero[:, 0])
    if suspect_line_nums.any():
        LOGGER.info("Suspicious scan lines: %s", str(suspect_line_nums))

    if calib_type == 2:
        return rad

    # Central wavenumber:
    cwnum = header['radtempcnv'][0, irchn, 0]
    if irchn == 0:
        cwnum = cwnum / 1.0e2
    else:
        cwnum = cwnum / 1.0e3

    bandcor_2 = header['radtempcnv'][0, irchn, 1] / 1e5
    bandcor_3 = header['radtempcnv'][0, irchn, 2] / 1e6

    # Count to radiance conversion:
    rad = k1_ * count * count + k2_ * count + k3_

    if calib_type == 2:
        return rad

    all_zero = np.logical_and(np.logical_and(np.equal(k1_, 0),
                                             np.equal(k2_, 0)),
                              np.equal(k3_, 0))
    idx = np.indices((all_zero.shape[0],))
    suspect_line_nums = np.repeat(idx[0], all_zero[:, 0])
    if suspect_line_nums.any():
        LOGGER.info("Suspect scan lines: %s", str(suspect_line_nums))

    ir_const_1 = 1.1910659e-5
    ir_const_2 = 1.438833

    t_planck = (ir_const_2 * cwnum) / \
        np.log(1 + ir_const_1 * cwnum * cwnum * cwnum / rad)

    # Band corrections applied to t_planck to get correct
    # brightness temperature for channel:
    if bandcor_2 < 0:  # Post AAPP-v4
        tb_ = bandcor_2 + bandcor_3 * t_planck
    else:  # AAPP 1 to 4
        tb_ = (t_planck - bandcor_2) / bandcor_3

    # tb_[tb_ <= 0] = np.nan
    # Data with count=0 are often related to erroneous (bad) lines, but in case
    # of saturation (channel 3b) count=0 can be observed and associated to a
    # real measurement. So we leave out this filtering to the user!
    # tb_[count == 0] = np.nan
    # tb_[rad == 0] = np.nan
    return np.ma.masked_array(tb_, np.isnan(tb_))


def show(data, negate=False):
    """Show the stetched data.
    """
    from PIL import Image
    data = np.array((data - data.min()) * 255.0 /
                    (data.max() - data.min()), np.uint8)
    if negate:
        data = 255 - data
    img = Image.fromarray(data)
    img.show()

CASES = {
    "avhrr": load_avhrr,
    "avhrr/3": load_avhrr,
}


def main():
    import sys
    from mpop.utils import debug_on

    debug_on()
    scene = AAPP1b(sys.argv[1])
    scene.read()
    for name, val in zip(scene._header[0].dtype.names, scene._header[0][0]):
        print(name + " " + str(val))
    starttime = datetime.datetime(scene._header[0][0]["startdatayr"],
                                  1, 1, 0, 0)
    starttime += \
        datetime.timedelta(days=int(scene._header[0][0]["startdatady"]) - 1,
                           seconds=scene._header[0][0]["startdatatime"] /
                           1000.0)
    print("starttime: " + str(starttime))
    endtime = datetime.datetime(scene._header[-1][0]["enddatayr"], 1, 1, 0, 0)
    endtime += \
        datetime.timedelta(days=int(scene._header[-1][0]["enddatady"]) - 1,
                           seconds=scene._header[-1][0]["enddatatime"] /
                           1000.0)
    print("endtime: " + str(endtime))
    # print scene._data['hrpt'].shape
    # show(scene._data['hrpt'][:, :, 4].astype(np.float))
    # raw_input()
    scene.calibrate()
    scene.navigate()
    for i__ in AVHRR_CHANNEL_NAMES:
        data_ = scene.channels[i__]
        print >> sys.stderr, "%-3s" % i__, \
            "%6.2f%%" % (100. * (float(np.ma.count(data_)) / data_.size)), \
            "%6.2f, %6.2f, %6.2f" % (data_.min(), data_.mean(), data_.max())
    show(scene.channels['2'], negate=False)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Helge Pfeiffer <rhp@dmi.dk>
#   Lars Orum Rasmussen <ras@dmi.dk>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Sentinel-1 reader
"""

import ConfigParser
import os

from mipp.read_geotiff import read_geotiff
from mipp.xsar import S1A

from datetime import datetime
import numpy as np

from glob import glob

from pyresample import geometry
import mpop
from mpop import CONFIG_PATH
import logging

LOG = logging.getLogger(__name__)

from mpop.plugin_base import Reader

class SentinelGRDChannel(mpop.channel.GenericChannel):

    def __init__(self, name='unknown', resolution='unknown'):
        mpop.channel.GenericChannel.__init__(self)
        self._is_loaded = False
        self.name = name
        self.resolution = resolution
        self.data = None
        self.shape = None
        self._projectables = []
        self._projectables.append(name)

    def is_loaded(self):
        return self._is_loaded

    def set_loaded(self):
        self._is_loaded = not self._is_loaded

    def project(self, coverage):
        """Project what can be projected in the product.
        """
        import copy
        res = copy.copy(self)
        res.data = coverage.project_array(self.data)
        return res

class GeoTiffReader(Reader):

    pformat = "mipp_sentinel"

    def load(self, satscene, **kwargs):

        LOG.debug('channels to load: ' + str(satscene.channels_to_load))
        conf = ConfigParser.ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
        options = {}

        for option, value in conf.items(satscene.instrument_name + "-level2",
                                        raw=True):
            options[option] = value

        options["resolution"] = kwargs.get("resolution", 'unknown')
        values = {"orbit": satscene.orbit}

        path_template = datetime.strftime(satscene.time_slot, options["dir"]) % values

        dirlist = glob(path_template)
        
        if len(dirlist) != 1:
            raise IOError("Couldn't identify a unique measurments directory!, " +
                          "from path_template '%s'" % path_template)

        dirname = dirlist[0]
        if not os.path.exists(dirname):
            raise IOError('Directory ' + str(dirname) + ' does not exist')

        # Read meta data
        mda = S1A.read_metadata(dirname)
        channels_available = set(mda.channels.keys())

        ##filelist = glob(os.path.join(dirname, options["filename"]))
        ##if len(filelist) == 0:
        ##    LOG.warning('No files found!')

        channels_to_load = satscene.channels_to_load.intersection(channels_available)
        
        # Loading of channels
        LOG.debug('available channels to load: ' + str(channels_to_load))
        for channel_name in channels_to_load:
            channel_file = mda.channels[channel_name]
            LOG.debug("Load channel: '%s' %s" % (channel_name, str(channel_file)))
            lons, lats, data = self.load_channel(channel_file)
    
            chn = SentinelGRDChannel(channel_name, mda.pixel_spacing[0])
            chn.area = geometry.SwathDefinition(lons=lons, lats=lats)
            chn.data = np.ma.masked_array(data)
            chn.shape = data.shape
            chn.set_loaded()
            satscene[channel_name] = chn

        satscene.info['manifest'] = mda


    def load_channel(self, filename):
        """Load one sentinel channel file"""
        from geotiepoints.basic_interpolator import BasicSatelliteInterpolator

        params, data = read_geotiff(filename)
        tie_lons = params['tiepoints']['lons']
        tie_lats = params['tiepoints']['lats']
        tie_cols = params['tiepoints']['cols']
        tie_rows = params['tiepoints']['rows']

        interpolator = BasicSatelliteInterpolator(tie_cols, 
                                                  tie_rows, 
                                                  tie_lats, 
                                                  tie_lons)
        lats, lons = interpolator.interpolate()
        
        return lons, lats, data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2012, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to EPS Avhrr/3 level 1b format.
http://oiswww.eumetsat.org/WEBOPS/eps-pg/AVHRR/AVHRR-PG-11L1bFormat.htm
http://www.eumetsat.int/idcplg?IdcService=GET_FILE&\
dDocName=PDF_TEN_96167-EPS-GPFS&RevisionSelectionMethod=LatestReleased
http://www.eumetsat.int/idcplg?IdcService=GET_FILE&\
dDocName=PDF_TEN_97231-EPS-AVHRR&RevisionSelectionMethod=LatestReleased
"""
import struct
import datetime
import numpy as np
import math
from scipy import interpolate
import os.path
import glob
from ConfigParser import ConfigParser
from mpop import CONFIG_PATH
import logging

LOG = logging.getLogger(__name__)

RECORD_CLASS = ["Reserved", "MPHR", "SPHR",
                "IPR", "GEADR", "GIADR",
                "VEADR", "VIADR", "MDR"]

INSTRUMENT_GROUP = ["GENERIC", "AMSU-A", "ASCAT", "ATOVS", "AVHRR/3", "GOME",
                    "GRAS", "HIRS/4", "IASA", "MHS", "SEM", "ADCS", "SBUV",
                    "DUMMY", "ARCHIVE", "IASI_L2"]

MPHR = [100, 100, 100, 100, 100, 37, 36, 36, 35, 36, 48, 48, 48, 48, 37, 38, 38,
        38, 38, 48, 48, 34, 34, 36, 48, 48, 38, 38, 44, 51, 44, 44, 44, 44, 44,
        44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
        35, 48, 39, 39, 39, 39, 39, 39, 39, 39, 39, 39, 39, 39, 39, 41, 41, 41,
        34]

SPHR = [38, 36]

GIADR_SUB = [None, "GIADR-RADIANCE", "GIADR-ANALOG"]

MAX_SCAN_LINES = 2000


def get_bit(bitstring, bit):
    """Get a given *bit* from *bitstring*.
    """
    return bitstring & (2 ** bit)


def read_u_bytes(fdes, size):
    """Read unsigned bytes, and scale it by 10 ** *sf_*.
    """
    cases = {
        1: ">B",
        2: ">H",
        4: ">I",
        8: ">Q"
    }
    return struct.unpack(cases[size], fdes.read(size))[0]


def read_bytes(fdes, size, sf_=0):
    """Read signed bytes, and scale it by 10 ** *sf_*.
    """
    cases = {
        1: ">b",
        2: ">h",
        4: ">i",
        8: ">q"
    }
    if sf_ != 0:
        return struct.unpack(cases[size], fdes.read(size))[0] * 10.0 ** sf_
    else:
        return struct.unpack(cases[size], fdes.read(size))[0]


def read_short_cds(fdes):
    """Read a short cds date.
    """

    difference = datetime.timedelta(days=read_u_bytes(fdes, 2),
                                    milliseconds=read_u_bytes(fdes, 4))
    epoch = datetime.datetime(2000, 1, 1)
    return epoch + difference


def read_long_cds(fdes):
    """Read a long cds date.
    """

    difference = datetime.timedelta(days=read_u_bytes(fdes, 2),
                                    milliseconds=read_u_bytes(fdes, 4),
                                    microseconds=read_u_bytes(fdes, 2))

    epoch = datetime.datetime(2000, 1, 1)
    return epoch + difference


def read_ascii_field(fdes, size):
    """Read an ascii field.
    """
    field_name = fdes.read(30).strip()
    fdes.read(2)
    field_value = fdes.read(size - 33).strip()
    fdes.read(1)
    return (field_name, field_value)


def read_bitstring(fdes, size):
    """Read a bit string.
    """
    cases = {
        1: ">B",
        2: ">H",
        4: ">I",
        8: ">Q"
    }
    return struct.unpack(cases[size], fdes.read(size))[0]


def print_bitstring(s__):
    """Print a bitstring.
    """

    res = ""
    ts_ = s__

    i = 0
    for i in range(16):
        res = str(ts_ & 1) + res
        ts_ = ts_ >> 1

    del i
    print res


def read_grh(fdes):
    """Read GRH.
    """

    grh = {}
    record_class = fdes.read(1)
    if record_class == "":
        return "EOF"
    grh["RECORD_CLASS"] = RECORD_CLASS[ord(record_class)]
    grh["INSTRUMENT_GROUP"] = INSTRUMENT_GROUP[ord(fdes.read(1))]
    grh["RECORD_SUBCLASS"] = ord(fdes.read(1))
    grh["RECORD_SUBCLASS_VERSION"] = ord(fdes.read(1))
    grh["RECORD_SIZE"] = read_u_bytes(fdes, 4)
    grh["RECORD_START_TIME"] = read_short_cds(fdes)
    grh["RECORD_STOP_TIME"] = read_short_cds(fdes)
    return grh


def read_mphr(fdes, grh, metadata):
    """Read MPHR.
    """

    del grh

    for i in MPHR:
        field_name, field_value = read_ascii_field(fdes, i)
        try:
            metadata[field_name] = eval(field_value)
        except:
            metadata[field_name] = str(field_value)

    return metadata


def read_sphr(fdes, grh, metadata):
    """Read SPHR.
    """

    del grh

    if(metadata["INSTRUMENT_ID"] != "AVHR"):
        raise NotImplementedError("Only Avhrr for now...")

    fdes.read(49)
    for i in SPHR:
        field_name, field_value = read_ascii_field(fdes, i)
        try:
            metadata[field_name] = eval(field_value)
        except:
            metadata[field_name] = str(field_value)


def read_ipr(fdes, grh, metadata):
    """Read IPR.
    """

    del grh, metadata

    ipr = {}
    ipr["TARGET_RECORD_CLASS"] = read_u_bytes(fdes, 1)
    ipr["TARGET_INSTRUMENT_GROUP"] = read_u_bytes(fdes, 1)
    ipr["TARGET_RECORD_SUBCLASS"] = read_u_bytes(fdes, 1)
    ipr["TARGET_RECORD_OFFSET"] = read_u_bytes(fdes, 4)
    return ipr


def read_geadr(fdes, grh, metadata):
    """Read GEADR.
    """

    del metadata

    geadr = {}
    field_name, field_val = read_ascii_field(fdes, grh["RECORD_SIZE"] - 20)
    geadr[field_name] = field_val
    return geadr


def read_giadr(fdes, grh, metadata):
    """Read GIADR.
    """
    if(metadata["INSTRUMENT_ID"] != "AVHR"):
        raise NotImplementedError("Only Avhrr for now...")
    if(metadata["PROCESSING_LEVEL"] != "1B"):
        raise NotImplementedError("Only level 1B for now...")

    if grh["RECORD_SUBCLASS"] == 1:
        return read_giadr_radiance(fdes, grh, metadata)
    elif grh["RECORD_SUBCLASS"] == 2:
        return read_giadr_analog(fdes, grh, metadata)
    elif grh["RECORD_SUBCLASS"] == 99:
        fdes.read(grh["RECORD_SIZE"] - 20)
        return
    else:
        raise ValueError("Undefined subclass " +
                         str(grh["RECORD_SUBCLASS"]) +
                         ", version " +
                         str(grh["RECORD_SUBCLASS_VERSION"]) +
                         "...")


def read_giadr_radiance(fdes, grh, metadata):
    """Read GIADR.
    """

    del grh
    del metadata

    giadr = {}
    giadr["RAMP_CALIBRATION_COEFFICIENT"] = read_bitstring(fdes, 2)
    giadr["YEAR_RECENT_CALIBRATION"] = read_u_bytes(fdes, 2)
    giadr["DAY_RECENT_CALIBRATION"] = read_u_bytes(fdes, 2)
    giadr["PRIMARY_CALIBRATION_ALGORITHM_ID"] = read_u_bytes(fdes, 2)
    giadr["PRIMARY_CALIBRATION_ALGORITHM_OPTION"] = read_bitstring(fdes, 2)
    giadr["SECONDARY_CALIBRATION_ALGORITHM_ID"] = read_u_bytes(fdes, 2)
    giadr["SECONDARY_CALIBRATION_ALGORITHM_OPTION"] = read_bitstring(fdes, 2)
    giadr["IR_TEMPERATURE1_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE1_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE1_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE1_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE1_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE1_COEFFICIENT6"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE2_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE2_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE2_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE2_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE2_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE2_COEFFICIENT6"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE3_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE3_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE3_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE3_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE3_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE3_COEFFICIENT6"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE4_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE4_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE4_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE4_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE4_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["IR_TEMPERATURE4_COEFFICIENT6"] = read_bytes(fdes, 2)
    giadr["CH1_SOLAR_FILTERED_IRRADIANCE"] = read_bytes(fdes, 2, -1)
    giadr["CH1_EQUIVALENT FILTER_WIDTH"] = read_bytes(fdes, 2, -3)
    giadr["CH2_SOLAR_FILTERED_IRRADIANCE"] = read_bytes(fdes, 2, -1)
    giadr["CH2_EQUIVALENT FILTER_WIDTH"] = read_bytes(fdes, 2, -3)
    giadr["CH3A_SOLAR_FILTERED_IRRADIANCE"] = read_bytes(fdes, 2, -1)
    giadr["CH3A_EQUIVALENT FILTER_WIDTH"] = read_bytes(fdes, 2, -3)
    giadr["CH3B_CENTRAL_WAVENUMBER"] = read_bytes(fdes, 4, -2)
    giadr["CH3B_CONSTANT1"] = read_bytes(fdes, 4, -5)
    giadr["CH3B_CONSTANT2_SLOPE"] = read_bytes(fdes, 4, -6)
    giadr["CH4_CENTRAL_WAVENUMBER"] = read_bytes(fdes, 4, -3)
    giadr["CH4_CONSTANT1"] = read_bytes(fdes, 4, -5)
    giadr["CH4_CONSTANT2_SLOPE"] = read_bytes(fdes, 4, -6)
    giadr["CH5_CENTRAL_WAVENUMBER"] = read_bytes(fdes, 4, -3)
    giadr["CH5_CONSTANT1"] = read_bytes(fdes, 4, -5)
    giadr["CH5_CONSTANT2_SLOPE"] = read_bytes(fdes, 4, -6)

    return giadr


def read_giadr_analog(fdes, grh, metadata):
    """Read GIADR.
    """
    giadr = {}

    giadr["PATCH_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_EXTENDED_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_EXTENDED_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_EXTENDED_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_EXTENDED_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["PATCH_TEMPERATURE_EXTENDED_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["PATCH_POWER_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["PATCH_POWER_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["PATCH_POWER_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["PATCH_POWER_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["PATCH_POWER_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["RADIATOR_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["RADIATOR_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["RADIATOR_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["RADIATOR_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["RADIATOR_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE1_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE1_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE1_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE1_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE1_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE2_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE2_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE2_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE2_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE2_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE3_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE3_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE3_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE3_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE3_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE4_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE4_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE4_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE4_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["BLACKBODY_TEMPERATURE4_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_CURRENT_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_CURRENT_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_CURRENT_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_CURRENT_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_CURRENT_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["MOTOR_CURRENT_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["MOTOR_CURRENT_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["MOTOR_CURRENT_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["MOTOR_CURRENT_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["MOTOR_CURRENT_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["EARTH_SHIELD_POSITION_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["EARTH_SHIELD_POSITION_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["EARTH_SHIELD_POSITION_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["EARTH_SHIELD_POSITION_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["EARTH_SHIELD_POSITION_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["ELECTRONIC_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["COOLER_HOUSING_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["COOLER_HOUSING_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["COOLER_HOUSING_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["COOLER_HOUSING_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["COOLER_HOUSING_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["BASEPLATE_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["BASEPLATE_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["BASEPLATE_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["BASEPLATE_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["BASEPLATE_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["MOTOR_HOUSING_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["MOTOR_HOUSING_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["MOTOR_HOUSING_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["MOTOR_HOUSING_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["MOTOR_HOUSING_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["AD_CONVERTER_TEMPERATURE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["AD_CONVERTER_TEMPERATURE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["AD_CONVERTER_TEMPERATURE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["AD_CONVERTER_TEMPERATURE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["AD_CONVERTER_TEMPERATURE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["DETECTOR4_BIAS_VOLTAGE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["DETECTOR4_BIAS_VOLTAGE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["DETECTOR4_BIAS_VOLTAGE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["DETECTOR4_BIAS_VOLTAGE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["DETECTOR4_BIAS_VOLTAGE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["DETECTOR5_BIAS_VOLTAGE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["DETECTOR5_BIAS_VOLTAGE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["DETECTOR5_BIAS_VOLTAGE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["DETECTOR5_BIAS_VOLTAGE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["DETECTOR5_BIAS_VOLTAGE_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["CH3B_BLACKBODY_VIEW_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["CH3B_BLACKBODY_VIEW_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["CH3B_BLACKBODY_VIEW_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["CH3B_BLACKBODY_VIEW_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["CH3B_BLACKBODY_VIEW_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["CH4_BLACKBODY_VIEW_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["CH4_BLACKBODY_VIEW_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["CH4_BLACKBODY_VIEW_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["CH4_BLACKBODY_VIEW_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["CH4_BLACKBODY_VIEW_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["CH5_BLACKBODY_VIEW_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["CH5_BLACKBODY_VIEW_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["CH5_BLACKBODY_VIEW_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["CH5_BLACKBODY_VIEW_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["CH5_BLACKBODY_VIEW_COEFFICIENT5"] = read_bytes(fdes, 2)
    giadr["REFERENCE_VOLTAGE_COEFFICIENT1"] = read_bytes(fdes, 2)
    giadr["REFERENCE_VOLTAGE_COEFFICIENT2"] = read_bytes(fdes, 2)
    giadr["REFERENCE_VOLTAGE_COEFFICIENT3"] = read_bytes(fdes, 2)
    giadr["REFERENCE_VOLTAGE_COEFFICIENT4"] = read_bytes(fdes, 2)
    giadr["REFERENCE_VOLTAGE_COEFFICIENT5"] = read_bytes(fdes, 2)

    return giadr


def read_veadr(fdes, grh, metadata):
    """Read VEADR.
    """
    veadr = {}
    field_name, field_val = read_ascii_field(fdes, grh["RECORD_SIZE"] - 20)
    veadr[field_name] = field_val
    return veadr


def read_mdr(fdes, grh, metadata):
    """Read MDR.
    """
    if grh["RECORD_SUBCLASS"] != 2:
        raise ValueError("Only l1b supported for now")

    return read_mdr_1b(fdes, grh, metadata)


def read_mdr_1b(fdes, grh, metadata):
    """Read MDR section, 1B type.
    """
    mdr = {}

    mdr["DEGRADED_INST_MDR"] = read_u_bytes(fdes, 1)
    mdr["DEGRADED_PROC_MDR"] = read_u_bytes(fdes, 1)
    mdr["EARTH_VIEWS_PER_SCANLINE"] = read_bytes(fdes, 2)
    scanlength = mdr["EARTH_VIEWS_PER_SCANLINE"]
    array = (np.fromfile(file=fdes, dtype=">i2", count=scanlength * 5) *
             10 ** -2)
    array = array.reshape(5, scanlength)
    array[2, :] *= 10 ** -2
    mdr["SCENE_RADIANCES"] = array

    # Channels 1, 2, 3a in units of W/(m^2.sr).
    # Channels 3b, 4, 5 in units of mW/(m^2.sr.cm^-1).
    # Channels 1, 2, 4 & 5 with scale factor = 2.
    # Channels 3a or 3b with scale factor = 4.

    mdr["TIME_ATTITUDE"] = read_u_bytes(fdes, 4)
    mdr["EULER_ANGLE"] = (read_bytes(fdes, 2), read_bytes(fdes, 2),
                          read_bytes(fdes, 2))
    mdr["NAVIGATION_STATUS"] = read_bitstring(fdes, 4)
    mdr["SPACECRAFT_ALTITUDE"] = read_u_bytes(fdes, 4)
    mdr["ANGULAR_RELATIONS_FIRST"] = (read_bytes(fdes, 2), read_bytes(fdes, 2),
                                      read_bytes(fdes, 2), read_bytes(fdes, 2))
    mdr["ANGULAR_RELATIONS_LAST"] = (read_bytes(fdes, 2), read_bytes(fdes, 2),
                                     read_bytes(fdes, 2), read_bytes(fdes, 2))
    mdr["EARTH_LOCATION_FIRST"] = np.array((read_bytes(fdes, 4, -4),
                                            read_bytes(fdes, 4, -4)))
    mdr["EARTH_LOCATION_LAST"] = np.array((read_bytes(fdes, 4, -4),
                                           read_bytes(fdes, 4, -4)))

    mdr["NUM_NAVIGATION_POINTS"] = read_bytes(fdes, 2)
    mdr["ANGULAR_RELATIONS"] = np.fromfile(file=fdes, dtype=">i2",
                                           count=412) * 10 ** -2

    mdr["EARTH_LOCATIONS"] = np.fromfile(file=fdes, dtype=">i4",
                                         count=206) * 10 ** -4

    mdr["QUALITY_INDICATOR"] = read_bitstring(fdes, 4)
    mdr["SCAN_LINE_QUALITY"] = read_bitstring(fdes, 4)
    mdr["CALIBRATION_QUALITY"] = (read_bitstring(fdes, 2),
                                  read_bitstring(fdes, 2),
                                  read_bitstring(fdes, 2))
    mdr["COUNT_ERROR_FRAME"] = read_u_bytes(fdes, 2)
    mdr["CH123A_CURVE_SLOPE1"] = (read_bytes(fdes, 4),
                                  read_bytes(fdes, 4),
                                  read_bytes(fdes, 4))
    mdr["CH123A_CURVE_INTERCEPT1"] = (read_bytes(fdes, 4),
                                      read_bytes(fdes, 4),
                                      read_bytes(fdes, 4))
    mdr["CH123A_CURVE_SLOPE2"] = (read_bytes(fdes, 4),
                                  read_bytes(fdes, 4),
                                  read_bytes(fdes, 4))
    mdr["CH123A_CURVE_INTERCEPT2"] = (read_bytes(fdes, 4),
                                      read_bytes(fdes, 4),
                                      read_bytes(fdes, 4))
    mdr["CH123A_CURVE_INTERCEPTION"] = (read_bytes(fdes, 4),
                                        read_bytes(fdes, 4),
                                        read_bytes(fdes, 4))
    mdr["CH123A_TEST_CURVE_SLOPE1"] = (read_bytes(fdes, 4),
                                       read_bytes(fdes, 4),
                                       read_bytes(fdes, 4))
    mdr["CH123A_TEST_CURVE_INTERCEPT1"] = (read_bytes(fdes, 4),
                                           read_bytes(fdes, 4),
                                           read_bytes(fdes, 4))
    mdr["CH123A_TEST_CURVE_SLOPE2"] = (read_bytes(fdes, 4),
                                       read_bytes(fdes, 4),
                                       read_bytes(fdes, 4))
    mdr["CH123A_TEST_CURVE_INTERCEPT2"] = (read_bytes(fdes, 4),
                                           read_bytes(fdes, 4),
                                           read_bytes(fdes, 4))
    mdr["CH123A_TEST_CURVE_INTERCEPTION"] = (read_bytes(fdes, 4),
                                             read_bytes(fdes, 4),
                                             read_bytes(fdes, 4))
    mdr["CH123A_PRELAUNCH_CURVE_SLOPE"] = (read_bytes(fdes, 4),
                                           read_bytes(fdes, 4),
                                           read_bytes(fdes, 4))
    mdr["CH123A_PRELAUNCH_CURVE_INTERCEPT1"] = (read_bytes(fdes, 4),
                                                read_bytes(fdes, 4),
                                                read_bytes(fdes, 4))
    mdr["CH123A_PRELAUNCH_CURVE_SLOPE2"] = (read_bytes(fdes, 4),
                                            read_bytes(fdes, 4),
                                            read_bytes(fdes, 4))
    mdr["CH123A_PRELAUNCH_CURVE_INTERCEPT2"] = (read_bytes(fdes, 4),
                                                read_bytes(fdes, 4),
                                                read_bytes(fdes, 4))
    mdr["CH123A_PRELAUNCH_CURVE_INTERCEPTION"] = (read_bytes(fdes, 4),
                                                  read_bytes(fdes, 4),
                                                  read_bytes(fdes, 4))
    mdr["CH3B45_SECOND_TERM"] = (read_bytes(fdes, 4),
                                 read_bytes(fdes, 4),
                                 read_bytes(fdes, 4))
    mdr["CH3B45_FIRST_TERM"] = (read_bytes(fdes, 4),
                                read_bytes(fdes, 4),
                                read_bytes(fdes, 4))
    mdr["CH3B45_ZEROTH_TERM"] = (read_bytes(fdes, 4),
                                 read_bytes(fdes, 4),
                                 read_bytes(fdes, 4))
    mdr["CH3B45_TEST_SECOND_TERM"] = (read_bytes(fdes, 4),
                                      read_bytes(fdes, 4),
                                      read_bytes(fdes, 4))
    mdr["CH3B45_TEST_FIRST_TERM"] = (read_bytes(fdes, 4),
                                     read_bytes(fdes, 4),
                                     read_bytes(fdes, 4))
    mdr["CH3B45_TEST_ZEROTH_TERM"] = (read_bytes(fdes, 4),
                                      read_bytes(fdes, 4),
                                      read_bytes(fdes, 4))
    mdr["CLOUD_INFORMATION"] = np.fromfile(file=fdes, dtype="<i2",
                                           count=scanlength)

    mdr["FRAME_SYNCHRONISATION"] = (read_u_bytes(fdes, 2),
                                    read_u_bytes(fdes, 2),
                                    read_u_bytes(fdes, 2),
                                    read_u_bytes(fdes, 2),
                                    read_u_bytes(fdes, 2),
                                    read_u_bytes(fdes, 2))
    mdr["FRAME_INDICATOR"] = (read_bitstring(fdes, 2), read_bitstring(fdes, 2))

    mdr["TIME_CODE"] = (read_bitstring(fdes, 2), read_bitstring(fdes, 2),
                        read_bitstring(fdes, 2), read_bitstring(fdes, 2))
    mdr["RAMP_CALIB"] = (read_bitstring(fdes, 2), read_bitstring(fdes, 2),
                         read_bitstring(fdes, 2), read_bitstring(fdes, 2),
                         read_bitstring(fdes, 2))
    mdr["INTERNAL_TARGET_TEMPERATURE_COUNT"] = (read_bitstring(fdes, 2),
                                                read_bitstring(fdes, 2),
                                                read_bitstring(fdes, 2))

    # Digital B telemetry
    mdr["INSTRUMENT_INVALID_WORD_FLAG"] = read_bitstring(fdes, 2)
    mdr["DIGITAL_B_DATA"] = read_bitstring(fdes, 2)

    # Analog housekeeping data
    mdr["INSTRUMENT_INVALID_ANALOG_WORD_FLAG"] = read_bitstring(fdes, 4)
    mdr["PATCH_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["PATCH_EXTENDED_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["PATCH_POWER"] = read_u_bytes(fdes, 2)
    mdr["RADIATOR_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["BLACKBODY_TEMPERATURE1"] = read_u_bytes(fdes, 2)
    mdr["BLACKBODY_TEMPERATURE2"] = read_u_bytes(fdes, 2)
    mdr["BLACKBODY_TEMPERATURE3"] = read_u_bytes(fdes, 2)
    mdr["BLACKBODY_TEMPERATURE4"] = read_u_bytes(fdes, 2)
    mdr["ELECTRONIC_CURRENT"] = read_u_bytes(fdes, 2)
    mdr["MOTOR_CURRENT"] = read_u_bytes(fdes, 2)
    mdr["EARTH_SHIELD_POSITION"] = read_u_bytes(fdes, 2)
    mdr["ELECTRONIC_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["COOLER_HOUSING_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["BASEPLATE_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["MOTOR_HOUSING_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["AD_CONVERTER_TEMPERATURE"] = read_u_bytes(fdes, 2)
    mdr["DETECTOR4_VOLTAGE"] = read_u_bytes(fdes, 2)
    mdr["DETECTOR5_VOLTAGE"] = read_u_bytes(fdes, 2)
    mdr["CH3_BLACKBODY_VIEW"] = read_u_bytes(fdes, 2)
    mdr["CH4_BLACKBODY_VIEW"] = read_u_bytes(fdes, 2)
    mdr["CH5_BLACKBODY_VIEW"] = read_u_bytes(fdes, 2)
    mdr["REFERENCE_VOLTAGE"] = read_u_bytes(fdes, 2)

    return mdr

C1 = 1.191062e-05  # mW/(m2*sr*cm-4)
C2 = 1.4387863  # K/cm-1


def to_bt(arr, wc_, a__, b__):
    """Convert to BT.
    """
    val = np.log(1 + (C1 * (wc_ ** 3) / arr))
    t_star = C2 * wc_ / val
    return a__ + b__ * t_star


def to_refl(arr, solar_flux):
    """Convert to reflectances.
    """
    return arr * math.pi * 100.0 / solar_flux


def calibrate(channels, info_giadr):
    """convert the radiances to reflectances and bts.
    """
    channels[0, :, :] = to_refl(channels[0, :, :],
                                info_giadr["CH1_SOLAR_FILTERED_IRRADIANCE"])
    channels[1, :, :] = to_refl(channels[1, :, :],
                                info_giadr["CH2_SOLAR_FILTERED_IRRADIANCE"])
    channels[2, :, :] = to_refl(channels[2, :, :],
                                info_giadr["CH3A_SOLAR_FILTERED_IRRADIANCE"])
    channels[3, :, :] = to_bt(channels[3, :, :],
                              info_giadr["CH3B_CENTRAL_WAVENUMBER"],
                              info_giadr["CH3B_CONSTANT1"],
                              info_giadr["CH3B_CONSTANT2_SLOPE"])
    channels[4, :, :] = to_bt(channels[4, :, :],
                              info_giadr["CH4_CENTRAL_WAVENUMBER"],
                              info_giadr["CH4_CONSTANT1"],
                              info_giadr["CH4_CONSTANT2_SLOPE"])
    channels[5, :, :] = to_bt(channels[5, :, :],
                              info_giadr["CH5_CENTRAL_WAVENUMBER"],
                              info_giadr["CH5_CONSTANT1"],
                              info_giadr["CH5_CONSTANT2_SLOPE"])


def read(fdes):
    """Read the entire file.
    """

    metadata = {}

    cnt = 0

    channels = None
    lons = None
    lats = None
    samples = None

    geo_samples = 0
    scanlength = 0

    g3a = False
    g3b = False

    while True:

        grh = read_grh(fdes)
        if grh == "EOF":
            break

        record_class = grh["RECORD_CLASS"]
        if record_class not in CASES:
            raise NotImplementedError

        res = CASES[grh["RECORD_CLASS"]](fdes, grh, metadata)

        if record_class == "SPHR":
            scanlength = metadata["EARTH_VIEWS_PER_SCANLINE"]
            llats = np.zeros((MAX_SCAN_LINES, scanlength))
            llons = np.zeros((MAX_SCAN_LINES, scanlength))

            channels = np.ma.ones((6, MAX_SCAN_LINES, scanlength),
                                  dtype=np.float) * np.infty
            geo_samples = np.round(
                scanlength / metadata["NAV_SAMPLE_RATE"]) + 3
            samples = np.zeros(geo_samples, dtype=np.intp)
            samples[1:-1] = np.arange(geo_samples - 2) * 20 + 5 - 1
            samples[-1] = scanlength - 1
            lats = np.zeros((MAX_SCAN_LINES, geo_samples))
            lons = np.zeros((MAX_SCAN_LINES, geo_samples))

        if record_class == "GIADR" and grh["RECORD_SUBCLASS"] == 1:
            info_giadr = res

        if record_class == "MDR":
            three_a = get_bit(res["FRAME_INDICATOR"][0], 0)
            if three_a:
                channels[0:3, cnt, :] = res["SCENE_RADIANCES"][0:3]
                channels[4:, cnt, :] = res["SCENE_RADIANCES"][3:]
                g3a = True
            else:
                channels[0:2, cnt, :] = res["SCENE_RADIANCES"][0:2]
                channels[3:, cnt, :] = res["SCENE_RADIANCES"][2:]
                g3b = True

            lats[cnt, 1:-1] = res["EARTH_LOCATIONS"]\
                [np.arange(geo_samples - 2) * 2]
            lats[cnt, 0] = res["EARTH_LOCATION_FIRST"][0]
            lats[cnt, -1] = res["EARTH_LOCATION_LAST"][0]
            lons[cnt, 1:-1] = res["EARTH_LOCATIONS"]\
                [np.arange(geo_samples - 2) * 2 + 1]
            lons[cnt, 0] = res["EARTH_LOCATION_FIRST"][1]
            lons[cnt, -1] = res["EARTH_LOCATION_LAST"][1]
            # unwraping datum shift line
            lons[cnt, :] = np.rad2deg(np.unwrap(np.deg2rad(lons[cnt, :])))

            xnew = np.arange(scanlength)
            tck = interpolate.splrep(samples, lats[cnt, :], s=0)
            llats[cnt, :] = interpolate.splev(xnew, tck, der=0)
            tck = interpolate.splrep(samples, lons[cnt, :], s=0)
            llons[cnt, :] = interpolate.splev(xnew, tck, der=0)

            cnt += 1

    channels = channels[:, :cnt, :]
    llats = llats[:cnt, :]
    llons = llons[:cnt, :]
    llons[llons > 180] -= 360
    llons[llons < -180] += 360

    calibrate(channels, info_giadr)
    return channels, llats, llons, g3a, g3b, metadata["ORBIT_START"]


CASES = {"MPHR": read_mphr,
         "SPHR": read_sphr,
         "IPR": read_ipr,
         "GEADR": read_geadr,
         "GIADR": read_giadr,
         "VEADR": read_veadr,
         "MDR": read_mdr}


EPSILON = 0.001


def load(satscene):
    """Read data from file and load it into *satscene*.
    """
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value
    LOAD_CASES[satscene.instrument_name](satscene, options)


def load_avhrr(satscene, options):
    """Read avhrr data from file and load it into *satscene*.
    """

    if "filename" not in options:
        raise IOError("No filename given, cannot load.")
    values = {"INSTRUMENT": satscene.instrument_name[:4].upper(),
              "FNAME": satscene.satname[0].upper() + satscene.number
              }
    filename = os.path.join(
        options["dir"],
        (satscene.time_slot.strftime(options["filename"]) % values))
    LOG.debug("Looking for file %s" % satscene.time_slot.strftime(filename))
    file_list = glob.glob(satscene.time_slot.strftime(filename))

    if len(file_list) > 1:
        raise IOError("More than one l1b file matching!")
    elif len(file_list) == 0:
        raise IOError("No l1b file matching!")

    try:
        fdes = open(file_list[0])
        channels, lats, lons, g3a, g3b, orbit = read(fdes)

    finally:
        fdes.close()

    channels = np.ma.masked_invalid(channels)

    satscene["1"] = channels[0, :, :]
    satscene["2"] = channels[1, :, :]
    satscene["4"] = channels[4, :, :]
    satscene["5"] = channels[5, :, :]
    if g3a:
        satscene["3A"] = channels[2, :, :]
    if g3b:
        satscene["3B"] = channels[3, :, :]

    print "Inside eps_avhrr.load_avhrr: orbit = ", orbit
    #satscene.orbit = str(int(orbit) + 1)
    satscene.orbit = str(int(orbit))

    try:
        from pyresample import geometry
        satscene.area = geometry.SwathDefinition(lons=lons, lats=lats)
    except ImportError:
        satscene.area = None
        satscene.lat = lats
        satscene.lon = lons


def get_lonlat(satscene, row, col):
    try:
        if (satscene.area is None and
                (satscene.lat is None or satscene.lon is None)):
            load(satscene)
    except AttributeError:
        load(satscene)
    try:
        return satscene.area.lons[row, col], satscene.area.lats[row, col]
    except AttributeError:
        return satscene.lon[row, col], satscene.lat[row, col]


def get_lat_lon(satscene, resolution):
    """Read lat and lon.
    """
    del resolution

    return LAT_LON_CASES[satscene.instrument_name](satscene, None)


def get_lat_lon_avhrr(satscene, options):
    """Read lat and lon.
    """
    del options

    return satscene.lat, satscene.lon


LAT_LON_CASES = {
    "avhrr": get_lat_lon_avhrr
}

LOAD_CASES = {
    "avhrr": load_avhrr
}


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module for testing the mpop.channel module.
"""

import unittest
from mpop.channel import GenericChannel, Channel
import string
import numpy as np

# epsilon
E = 0.0001

class TestGenericChannel(unittest.TestCase):
    """Class for testing the GenericChannel class.
    """
    chan = None
    chan2 = None
    
    def test_init(self):
        """Creation of a generic channel.
        """
        self.chan = GenericChannel(name="newchan")
        self.assertEqual(self.chan.name, "newchan")
        
        numb = int(np.random.uniform(100000))
        self.assertRaises(TypeError, GenericChannel, name=numb)
        
        self.chan = GenericChannel()
        self.assertTrue(self.chan.name is None)
        
    def test_cmp(self):
        """Comparison of generic channels.
        """
        
        self.chan = GenericChannel(name = "newchan")
        self.chan2 = GenericChannel(name = "mychan")

        self.assertTrue(self.chan > self.chan2)

        self.chan = GenericChannel(name = "newchan")
        self.chan2 = "mychan"
        
        self.assertTrue(self.chan > self.chan2)

        self.chan = GenericChannel(name = "newchan")
        self.chan2 = GenericChannel(name = "newchan")

        self.assert_(self.chan == self.chan2)

        self.chan = GenericChannel()
        self.chan2 = GenericChannel(name = "newchan")

        self.assert_(self.chan < self.chan2)

        self.chan = GenericChannel(name = "newchan")
        self.chan2 = GenericChannel(name = "_mychan")

        self.assert_(self.chan < self.chan2)

        self.chan = GenericChannel(name = "_newchan")
        self.chan2 = GenericChannel(name = "mychan")

        self.assert_(self.chan > self.chan2)

    def test_area(self):
        """Area setting and retrieving.
        """
        self.chan = GenericChannel(name = "newchan")
        self.chan.area = "bla"
        self.assert_(self.chan.area == "bla")

        self.chan.area = None
        self.assert_(self.chan.area == None)

        class DummyArea(object):
            def __init__(self, area_extent, x_size, y_size, proj_id, proj_dict):
                self.area_extent = area_extent
                self.x_size = x_size
                self.y_size = y_size
                self.proj_id = proj_id
                self.proj_dict = proj_dict

        self.chan.area = DummyArea(1, 2, 3, 4, 5)
        self.assert_(self.chan.area.area_extent == 1)

        class DummyArea(object):
            def __init__(self, lons, lats):
                self.lons = lons
                self.lats = lats

        self.chan.area = DummyArea(1, 2)
        self.assert_(self.chan.area.lats == 2)

        self.assertRaises(TypeError, setattr, self.chan, "area", 1)


class TestChannel(unittest.TestCase):
    """Class for testing the Channel class.
    """
    chan = None
    chan2 = None
    
    def test_init(self):
        """Creation of a channel.
        """
        self.assertRaises(ValueError, Channel)

        # Name
        self.chan = Channel(name = "newchan")
        self.assertEqual(self.chan.name, "newchan")
        self.assertEqual(self.chan.wavelength_range,
                         [-np.inf, -np.inf, -np.inf])
        self.assertEqual(self.chan.resolution, 0)
        self.assert_(self.chan.data is None)

        numb = int(np.random.uniform(100000))
        self.assertRaises(TypeError, Channel, name = numb)
        numb = np.random.uniform() * 100000
        self.assertRaises(TypeError, Channel, name = numb)

        # Resolution
        numb = int(np.random.uniform(100000))
        self.assertRaises(ValueError, Channel, resolution = numb)

        numb = int(np.random.uniform(100000))
        self.chan = Channel(name = "newchan", resolution = numb)
        self.assertEqual(self.chan.name, "newchan")
        self.assertEqual(self.chan.wavelength_range,
                         [-np.inf, -np.inf, -np.inf])
        self.assertEqual(self.chan.resolution, numb)
        self.assert_(self.chan.data is None)

        self.assertRaises(TypeError, Channel,
                          name = "newchan",
                          resolution = "a")
        
        # Wavelength

        numbs = [np.random.uniform(100),
                 np.random.uniform(100),
                 np.random.uniform(100)]
        numbs.sort()

        self.chan = Channel(wavelength_range = numbs)
        self.assertEqual(self.chan.name, None)
        self.assertEqual(self.chan.wavelength_range, numbs)
        self.assertEqual(self.chan.resolution, 0)
        self.assert_(self.chan.data is None)

        self.assertRaises(TypeError, Channel,
                          wavelength_range = numbs[0:1])

        numbs.sort(reverse = True)
        self.assertRaises(ValueError, Channel,
                          wavelength_range = numbs)

        numbs = [int(np.random.uniform(100)),
                 int(np.random.uniform(100)),
                 int(np.random.uniform(100))]
        numbs.sort()        

        self.assertRaises(TypeError, Channel,
                          wavelength_range = numbs)

        self.assertRaises(TypeError, Channel,
                          wavelength_range = random_string(4))

        numb = np.random.uniform(100000)
        self.assertRaises(TypeError, Channel,
                          wavelength_range = numb)
        
        numb = int(np.random.uniform(100000))
        self.assertRaises(TypeError, Channel,
                          wavelength_range = numb)


        # Data

        data = np.random.rand(3, 3)
        
        self.assertRaises(ValueError, Channel, data = data)

        self.chan = Channel(name = "newchan", data = data)
        self.assertEqual(self.chan.name, "newchan")
        self.assertEqual(self.chan.wavelength_range,
                         [-np.inf, -np.inf, -np.inf])
        self.assertEqual(self.chan.resolution, 0)
        self.assert_(np.all(self.chan.data == data))

        mask = np.array(np.random.rand(3, 3) * 2, dtype = int)
        data = np.ma.array(data, mask = mask)
        
        self.chan = Channel(name = "newchan", data = data)
        self.assertEqual(self.chan.name, "newchan")
        self.assertEqual(self.chan.wavelength_range,
                         [-np.inf, -np.inf, -np.inf])
        self.assertEqual(self.chan.resolution, 0)
        self.assert_(np.all(self.chan.data == data))

        self.assertRaises(TypeError,
                          Channel,
                          name = "newchan",
                          data = random_string(4))

        numb = np.random.uniform(100000)
        self.assertRaises(TypeError,
                          Channel,
                          name = "newchan",
                          data = numb)
        
        numb = int(np.random.uniform(100000))
        self.assertRaises(TypeError,
                          Channel,
                          name = "newchan",
                          data = numb)
        
        numbs = [np.random.uniform(100),
                 np.random.uniform(100),
                 np.random.uniform(100)]
        self.assertRaises(TypeError,
                          Channel,
                          name = "newchan",
                          data = numbs)

    def test_cmp(self):
        """Comparison of channels.
        """
        
        self.chan = Channel(name = "newchan")
        self.chan2 = Channel(name = "mychan")

        self.assertTrue(self.chan > self.chan2)

        self.chan = Channel(name = "newchan")
        self.chan2 = "mychan"
        
        self.assertTrue(self.chan > self.chan2)

        self.chan = Channel(name = "newchan")
        self.chan2 = Channel(name = "newchan")

        self.assert_(self.chan == self.chan2)

        self.chan = Channel(wavelength_range=(1., 2., 3.))
        self.chan2 = Channel(name = "newchan")

        self.assert_(self.chan < self.chan2)

        self.chan = Channel(name = "newchan")
        self.chan2 = Channel(name = "_mychan")

        self.assert_(self.chan < self.chan2)

        self.chan = Channel(name = "_newchan")
        self.chan2 = Channel(name = "mychan")

        self.assert_(self.chan > self.chan2)

        self.chan = Channel(name = random_string(4),
                            wavelength_range = (1., 2., 3.))
        self.chan2 = Channel(name = random_string(4),
                             wavelength_range = (4., 5., 6.))

        self.assert_(self.chan < self.chan2)

        self.chan = Channel(name = "_" + random_string(4),
                            wavelength_range = (1., 2., 3.))
        self.chan2 = Channel(name = random_string(4),
                             wavelength_range = (4., 5., 6.))

        self.assert_(self.chan > self.chan2)


    def test_str(self):
        """String output for a channel.
        """
        self.chan = Channel(name="newchan",
                            wavelength_range=(1., 2., 3.),
                            resolution=1000)
        self.assertEqual(str(self.chan),
                         "'newchan: (1.000,2.000,3.000)μm, resolution 1000m,"
                         " not loaded'")

        self.chan.data = np.random.rand(3, 3)

        
        self.assertEqual(str(self.chan),
                         "'newchan: (1.000,2.000,3.000)μm, "
                         "shape (3, 3), "
                         "resolution 1000m'")
        

    def test_is_loaded(self):
        """Check load status of a channel.
        """
        data = np.random.rand(3, 3)
        
        self.chan = Channel(name = "newchan")
        self.assert_(not self.chan.is_loaded())

        self.chan = Channel(name = "newchan", data = data)
        self.assert_(self.chan.is_loaded())

    def test_as_image(self):
        """Check the geo_image version of the channel.
        """
        data = np.random.rand(3, 3)
        
        self.chan = Channel(name="newchan", data=data)
        img = self.chan.as_image(False)
        self.assert_(np.allclose(img.channels[0], data))
        self.assertEqual(img.mode, "L")
        img = self.chan.as_image(True)
        self.assertEqual(img.channels[0].max(), 1)
        self.assertEqual(img.channels[0].min(), 0)


    def test_check_range(self):
        """Check the range of a channel.
        """

        self.chan = Channel(name = "newchan")
        self.assertRaises(ValueError, self.chan.check_range)

        numb = np.random.uniform(10) 
        self.assertRaises(ValueError, self.chan.check_range, numb)

        # ndarray

        data = np.random.rand(3, 3)
        self.chan = Channel(name = "newchan", data = data)

        min_range = (data.max() - data.min()) / 2
        self.assert_(np.all(data == self.chan.check_range(min_range)))

        zeros = np.zeros_like(data)
        min_range = (data.max() - data.min()) + E
        self.assert_(np.all(zeros == self.chan.check_range(min_range)))

        # masked array

        mask = np.array(np.random.rand(3, 3) * 2, dtype = int)
        mask[1, 1] = False
        data = np.ma.array(data, mask = mask)
        self.chan = Channel(name = "newchan", data = data)

        min_range = (data.max() - data.min()) / 2
        self.assert_(np.all(data == self.chan.check_range(min_range)))
        self.assertEquals(data.count(),
                          self.chan.check_range(min_range).count())
        
        zeros = np.zeros_like(data)
        min_range = (data.max() - data.min()) + E
        self.assert_(np.all(zeros == self.chan.check_range(min_range)))

        data = np.ma.array(data, mask = True)
        self.chan = Channel(name = "newchan", data = data)
        self.assertEquals(0,
                          self.chan.check_range(min_range).count())
        self.assertEquals(data.count(),
                          self.chan.check_range(min_range).count())

        # Wrong type arguments

        self.assertRaises(TypeError, self.chan.check_range, random_string(4))

        self.assertRaises(TypeError,
                          self.chan.check_range,
                          [np.random.uniform()])


    def test_sunzen_corr(self):
        '''Test Sun zenith angle correction.
        '''

        import datetime as dt

        chan = Channel(name='test')
        
        original_value = 10.

        chan.data = original_value * np.ones((2,11))
        lats = np.zeros((2,11)) # equator
        lons = np.array([np.linspace(-90, 90, 11), np.linspace(-90, 90, 11)])

        # Equinox, so the Sun is at the equator
        time_slot = dt.datetime(2014,3,20,16,57)

        new_ch = chan.sunzen_corr(time_slot, lonlats=(lons, lats), limit=80.)

        # Test minimum after correction, accuracy of three decimals is enough
        #self.assertTrue(np.abs(10.000 - np.min(new_ch.data)) < 10**-3)
        self.assertAlmostEqual(10.000, np.min(new_ch.data), places=3)
        # Test maximum after correction
        self.assertAlmostEqual(57.588, np.max(new_ch.data), places=3)

        # There should be ten values at zenith angle >= 80 deg, and
        # these are all equal
        self.assertTrue(np.where(new_ch.data == \
                                     np.max(new_ch.data))[0].shape[0] == 10)

        # All values should be larger than the starting values
        self.assertTrue(np.all(new_ch.data > original_value))

        # Channel name
        self.assertEqual(new_ch.name, chan.name+'_SZC')
        
        # Test channel name in the info dict
        self.assertEqual(new_ch.name, chan.info['sun_zen_corrected'])

        # Test with several locations and arbitrary data
        chan = Channel(name='test2')
        chan.data = np.array([[0., 67.31614275, 49.96271995,
                               99.41046645, 29.08660989],
                              [87.61007584, 79.6683524, 53.20397351,
                               29.88260374, 62.33623915],
                              [60.49283004, 54.04267222, 32.72365906,
                               91.44995651, 32.27232955],
                              [63.71580638, 69.57673795, 7.63064373,
                               32.15683105, 9.05786335],
                              [65.61434337, 33.2317155, 18.77672384,
                               30.13527574, 23.22572904]])
        lons = np.array([[116.28695847, 164.1125604, 40.77223701,
                          -113.54699788, 133.15558442],
                         [-17.18990601, 75.17472034, 12.81618371,
                           -40.75524952, 40.70898002],
                         [42.74662341, 164.05671859, -166.58469404,
                          -58.16684483, -144.97963063],
                         [46.26303645, -167.48682034, 170.28131412,
                          -17.80502488, -63.9031154],
                         [-107.14829679, -147.66665952, -0.75970554,
                           77.701768, -130.48677807]])
        lats = np.array([[-51.53681682, -83.21762788, 5.91008672, 
                           22.51730385, 66.83356427],
                         [82.78543163,  23.1529456 ,  -7.16337152,
                          -68.23118425, 28.72194953],
                         [31.03440852, 70.55322517, -83.61780288,
                          29.88413938, 25.7214828],
                         [-19.02517922, -19.20958728, -14.7825735,
                           22.66967876, 67.6089238],
                         [45.12202477, 61.79674149, 58.71037615,
                          -62.04350423, 13.06405864]])
        time_slot = dt.datetime(1998, 8, 1, 10, 0)

        # These are the expected results
        results = np.array([[0., 387.65821593, 51.74080022,
                             572.48205988, 138.96586013],
                            [227.24857818, 105.53045776, 62.24134162,
                             172.0870564, 64.12902666],
                            [63.08646652, 311.21934562, 188.44804188,
                             526.63931022, 185.84893885],
                            [82.86856236, 400.6764648, 43.9431259,
                             46.58056343, 36.04457644],
                            [377.85794388, 191.3738223, 27.55002934,
                             173.54213642, 133.75164285]])

        new_ch = chan.sunzen_corr(time_slot, lonlats=(lons, lats), limit=80.)
        self.assertAlmostEqual(np.max(results-new_ch.data), 0.000, places=3)
        
#    def test_project(self):
#        """Project a channel.
#        """
        # from pp.coverage import SatProjCov
#         from pp.scene import SatelliteInstrumentScene
        
#         cov = SatProjCov(SatelliteInstrumentScene(area = "euro"),
#                          "scan", 1000)
#         data = np.tile(np.array([[1, 2],[3, 4]]), (256, 256))
#         self.chan = Channel(name = "newchan", data = data)
#         self.chan.project(cov)

def random_string(length, choices=string.letters):
    """Generates a random string with elements from *set* of the specified
    *length*.
    """
    import random
    return "".join([random.choice(choices)
                    for i in range(length)])


def suite():
    """The test suite for test_channel.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestGenericChannel))
    mysuite.addTest(loader.loadTestsFromTestCase(TestChannel))
    
    return mysuite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the module :mod:`pp.satellites`.
"""
import ConfigParser
import random
import unittest

import mpop.instruments.visir
import mpop.satellites

INSTRUMENTS = ()

def random_string(length,
                  choices="abcdefghijklmnopqrstuvwxyz"
                  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
    """Generates a random string with elements from *set* of the specified
    *length*.
    """
    return "".join([random.choice(choices)
                    for dummy_itr in range(length)])


def patch_configparser():
    """Patch to fake ConfigParser.
    """
    class FakeConfigParser:
        """Dummy ConfigParser class.
        """
        def __init__(self, *args, **kwargs):
            pass
        
        def sections(self, *args, **kwargs):
            """Dummy sections method.
            """
            self = self
            del args, kwargs
            sections = []
            for i in INSTRUMENTS:
                for j in range(int(random.random() * 10 + 1)):
                    sections += [i + str(j)]
            return sections
        
        def read(self, *args, **kwargs):
            """Dummy read method
            """
            del args, kwargs
            self = self

        def get(self, *args, **kwargs):
            """Dummy get method
            """
            del kwargs
            self = self
            if args[1] == "instruments":
                return str(INSTRUMENTS)
            if args[1] == "name":
                return "'" + random_string(3) + "'"
            if args[1] == "resolution":
                return str(int(random.random() * 50000 + 1))
            if args[1] == "frequency":
                return str(random.random())
            if args[1] == "module":
                return random_string(8)
            
    ConfigParser.OldConfigParser = ConfigParser.ConfigParser
    ConfigParser.ConfigParser = FakeConfigParser

def unpatch_configparser():
    """Unpatch fake ConfigParser.
    """
    ConfigParser.ConfigParser = ConfigParser.OldConfigParser
    delattr(ConfigParser, "OldConfigParser")

def patch_scene():
    """Patch the :mod:`mpop.instruments.visir` module to avoid using it in these
    tests.
    """
    class FakeChannel(object):
        """FakeChannel class.
        """
        def __init__(self, val):
            self.data = val

        def check_range(self, *args):
            """Dummy check_range function.
            """
            del args
            return self.data

    
    class FakeSatscene(object):
        """Fake SatelliteInstrumentScene.
        """
        __version__ = "fake"
        def __init__(self):
            self.channels = None
            self.area = None
            self.time_slot = None
            self.error = []
        
        def check_channels(self, *args):
            """Dummy check_channels function.
            """
            self.channels = args

        def __getitem__(self, key):
            if key in self.error:
                raise KeyError()
            return FakeChannel(key)
    mpop.instruments.visir.OldVisirCompositer = mpop.instruments.visir.VisirCompositer
    mpop.instruments.visir.VisirCompositer = FakeSatscene
    reload(mpop.satellites)
    

def unpatch_scene():
    """Unpatch the :mod:`mpop.scene` module.
    """
    mpop.instruments.visir.VisirCompositer = mpop.instruments.visir.OldVisirCompositer
    delattr(mpop.instruments.visir, "OldVisirCompositer")
    reload(mpop)
    reload(mpop.instruments)
    reload(mpop.instruments.visir)
    reload(mpop.satellites)


class TestSatellites(unittest.TestCase):
    """Test the satellites base functions.
    """

    def setUp(self):
        """Patch stuff.
        """
        patch_configparser()
        patch_scene()
        
    def test_buildinstrument(self):
        """Test the :func:`mpop.satellites.build_instrument` function.
        """
        name = random_string(10)
        #ch_list = [random_string(10), random_string(12)]
        inst = mpop.satellites.build_instrument_compositer(name)

        # Test that the patches are applied
        self.assertEquals(inst.__version__, "fake")

        
        #self.assertEquals(inst.channel_list, ch_list)
        self.assertEquals(inst.instrument_name, name)
        self.assertEquals(inst.mro()[1], mpop.instruments.visir.VisirCompositer)

    def test_build_satellite_class(self):
        """Test the :func:`mpop.satellites.build_satellite_class` function.
        """
        global INSTRUMENTS
        inst = random_string(10)
        INSTRUMENTS = (inst, )
        satname = random_string(10)
        satnumber = random_string(10)
        satvar = random_string(10)
        myclass = mpop.satellites.build_sat_instr_compositer((satname,
                                                              satnumber,
                                                              satvar),
                                                             inst)
        #self.assertEquals(myclass.satname, satname)
        #self.assertEquals(myclass.number, satnumber)
        #self.assertEquals(myclass.variant, satvar)
        self.assertEquals(myclass.mro()[1].__name__,
                          inst.capitalize() +
                          "Compositer")
            
    def test_get_satellite_class(self):
        """Test the :func:`mpop.satellites.get_satellite_class` function.
        """
        global INSTRUMENTS

        inst = random_string(10)
        INSTRUMENTS = ("avhrr", inst)
        satname = random_string(11)
        satnumber = random_string(10)
        satvar = random_string(10)
        klass = mpop.satellites.get_sat_instr_compositer((satname,
                                                          satnumber,
                                                          satvar),
                                                         inst)
        self.assertTrue(klass.mro()[0].__name__.startswith(
            satvar.capitalize() + satname.capitalize() +
            satnumber.capitalize()))


        INSTRUMENTS = (inst,)
        satname = random_string(11)
        satnumber = random_string(10)
        satvar = random_string(10)
        klass = mpop.satellites.get_sat_instr_compositer((satname,
                                                          satnumber,
                                                          satvar),
                                                         inst)
        pklass = klass.mro()[0]
        self.assertTrue(pklass.__name__.startswith(
            satvar.capitalize() + satname.capitalize() +
            satnumber.capitalize()))

    def tearDown(self):
        """Unpatch stuff.
        """
        unpatch_configparser()
        unpatch_scene()

def suite():
    """The test suite for test_satellites.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestSatellites))
    
    return mysuite
